# -*- coding: utf-8 -*-
"""Module containing the server profile class

The class :class:`ServerProfile` is the base for storing TLS server scan results.
A basic work flow for a server scan is as follows:

* The server profile is initialized and contains no data at the beginning.

* The scan is started, and the workers do their job. They are storing their
  results in the server profile. For example, first the supported TLS versions
  are determined together with the supported cipher suites. Next, the workers
  scan for the supported groups and the supported signature algorithms.

* At this point in time the server profile can be read by subsequent workers. E.g.,
  DH-parameters are only checked, if the server profile contains cipher suites
  using the DH key exchange. The DH-scanner worker will use the corresponding
  TLS versions, cipher suites and other parameters, knowing that they are all
  supported by the server. Then the worker will store its results in the server
  profile as well.

* After the scan is finished, the server profile can either be stored in a JSON/YAML
  file, or an extract from the server profile can be displayed on the screen (by a
  dedicated worker).

The server profile can also be used as a base for customized test cases: if a
scan has been executed against the target and the server profile has been serialized
to a file, then this server profile can be deserialized, and thus the test case can
use the data from it to evaluate if the feature to be tested is supported by the
target, and if so, the server profile values can be used to initiate connection to
the target with parameters known to be supported. This is exactly what the provided
workers do.

The YAML/JSON schema of a serialized server profile is described by the class
:class:`ServerProfileSchema` (and its nested schema classes). Look at the code
if you want to see the detailed structure.
"""
# import basic stuff
import abc
import logging
from typing import List, Tuple, Type, Any, Dict, Optional

# import own stuff
import tlsmate.cert as crt
import tlsmate.cert_chain as cert_chain
import tlsmate.mappings as mappings
import tlsmate.pdu as pdu
import tlsmate.structs as structs
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff
import yaml
from cryptography.hazmat.primitives.asymmetric import rsa, ed25519, ed448, dsa, ec
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography import x509
from marshmallow import fields, Schema, post_load, post_dump, pre_dump, INCLUDE
from marshmallow_oneofschema import OneOfSchema  # type: ignore

# #### Helper classes


class _YamlBlockStyle(str):
    """Class used to indicate that a string shall be serialized using the block style.
    """


def _literal_presenter(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")


yaml.add_representer(_YamlBlockStyle, _literal_presenter)


class FieldsEnumString(fields.String):
    """Adapted marshmallow field for ExtendedEnums.

    An extended enum is serialized to a string field, containing the name of the enum.

    Example:
        ::

            In the schema class:
                field = FieldsEnumString(enum_class=tls.Version)

            In the data object:
                obj.field = tls.Version.TLS10

            Serialized:
                "field": "TLS10"

    Arguments:
        enum_class: The enum class used for deserialization.
    """

    default_error_messages = fields.String.default_error_messages
    default_error_messages.update({"not_in_enum": "String not defined in enum"})

    def __init__(self, enum_class: Type[tls.ExtendedEnum], **kwargs: Any) -> None:
        if enum_class is None:
            raise ValueError("FieldsEnumString: enum_class not given")

        # TODO: replace issubclass with isinstance?
        if not issubclass(enum_class, tls.ExtendedEnum):  # type: ignore
            raise ValueError("FieldsEnumString: class must be ExtendedEnum")

        self.enum_class = enum_class
        super().__init__(**kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        if not isinstance(value, self.enum_class):
            return None

        return super()._serialize(value.name, attr, obj, **kwargs)

    def _deserialize(self, value, attr, data, **kwargs):
        ret = self.enum_class.str2enum(value)
        if ret is None:
            raise self.make_error("not_in_enum")

        return ret


class FieldsBytes(fields.String):
    """Adapted marshmallow field for bytes.

    A bytes object will be serialized to a string where the bytes are separated by
    a colon.

    Example:
        ::

            b"Hallo" will be serialized to "48:61:6c:6c:6f"
    """

    default_error_messages = fields.String.default_error_messages
    default_error_messages.update({"not_in_enum": "String not defined in enum"})

    def _serialize(self, value, attr, obj, **kwargs):
        if not isinstance(value, (bytes, bytearray)):
            return None

        return pdu.string(value)

    def _deserialize(self, value, attr, data, **kwargs):
        return bytes.fromhex(
            "".join([c for c in value if c in "0123456789abcdefABCDEF"])
        )


class FieldsBlockString(fields.String):
    """Adapted marshmallow field strings which shall be displayed in block style (yaml).
    """

    def _serialize(self, value, attr, obj, **kwargs):
        return _YamlBlockStyle(value)


class ProfileSchema(Schema):
    """Wrapper class for easier deserialization to objects
    """

    _augments: List[Tuple[Type[Schema], Type[Schema]]] = []

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # self.unknown = INCLUDE
        self.unknown = INCLUDE

    @post_load
    def deserialize(
        self, data: Dict[str, Any], reuse_object: Optional[bool] = None, **kwargs
    ) -> Any:
        """Instantiate an object and define the properties according to the given dict.

        Arguments:
            data (dict): dict representing the properties of the object we need to
                deserialize.

        Returns:
            Deserialized data, either a newly instantiated object (if given by the
            __profile_class__ class property), or the original dict.
        """

        cls = getattr(self, "__profile_class__", None)
        if cls is None:
            return data

        cls_data = self._get_schema_data(data)
        if reuse_object:
            obj = reuse_object
            # TODO: avoid calling __init__
            obj.__init__(data=cls_data)  # type: ignore

        else:
            obj = cls(data=cls_data)

        for base_cls, ext_cls in self._augments:
            if base_cls is self.__class__:
                cls_data = ext_cls._get_schema_data(data)  # type: ignore
                ext_data = ext_cls().load(cls_data)
                for key, val in ext_data.items():
                    setattr(obj, key, val)

        if data:
            fields = ", ".join(data.keys())
            raise ValueError(
                f"fields not defined in schema {self.__class__.__name__}: {fields}"
            )

        return obj

    @post_dump(pass_original=True)
    def serialize(
        self, data: Dict[str, Any], orig: Any, **kwargs: Any
    ) -> Dict[str, Any]:
        """Deserialize properties not covered by the schema as well.

        Arguments:
            data: the serialized data so far
            orig: the original object subject to serialization

        Returns:
            the final dict representing the serialized object
        """

        for base_cls, ext_cls in self._augments:
            if base_cls is self.__class__:
                data.update(ext_cls().dump(orig))
        return data

    @classmethod
    def _get_schema_data(cls, data):
        cls_data = {}
        for key in cls._declared_fields.keys():
            if key in data:
                cls_data[key] = data[key]
                del data[key]
        return cls_data

    @classmethod
    def augment(cls, ext_cls: Type["ProfileSchema"]) -> Type["ProfileSchema"]:
        """Decorator to register scheme extensions.

        cls is the class to be extended.

        Examlple:
            @SPServerProfile.augment
            class MyExtensions(ProfileSchema):
                client_simulation = fields.Nested(...)
                a_number = fields.Integer()

        This extends the schema class SPServerProfile by two more attributes.
        Note, that __profile_class__ must not be given in the extension class. The
        name of the extension class has no relevance.

        Arguments:
            ext_cls: the schema class containing the extensions

        Returns:
            the class used to extend the base_cls class
        """

        cls._augments.append((cls, ext_cls))
        return ext_cls


class ProfileEnumSchema(Schema):
    """Wrapper class for simpler (de)serialization of Enums.
    """

    id = fields.Integer()
    name = fields.String()

    @post_load
    def deserialize(self, data: Dict[str, Any], **kwargs: Any) -> Any:
        # TODO: enable type check
        return self.__profile_class__(data["id"])  # type: ignore

    @pre_dump
    def serialize(self, obj: tls.ExtendedEnum, **kwargs: Any) -> Dict[str, Any]:
        return {"id": obj.value, "name": obj.name}


class SPObject(metaclass=abc.ABCMeta):
    """Basic class for profile data, provides easy deserialization from dicts.

    Arguments:
        data(dict): a dictionary which is converted into object properties.
        obj: an object to initialize the instance from.
    """

    def __init__(self, data: Optional[Dict[str, Any]] = None, **kwargs: Any) -> None:
        if data is None:
            data = {}

        for key, val in {**data, **kwargs}.items():
            if val is not None:
                setattr(self, key, val)


# ### Classes used for the server profile


class SPScanInfo(SPObject):
    """Data class for scan infos.
    """


class SPScanInfoSchema(ProfileSchema):
    """Schema for scan infos.
    """

    __profile_class__ = SPScanInfo
    command = fields.String()
    proxy = fields.String()
    run_time = fields.Float()
    start_date = fields.DateTime()
    start_timestamp = fields.Float()
    stop_date = fields.DateTime()
    stop_timestamp = fields.Float()
    version = fields.String()


class SPCompressionEnumSchema(ProfileEnumSchema):
    """Schema for compression method (enum)
    """

    __profile_class__ = tls.CompressionMethod


class SPGrease(SPObject):
    """Data class for GREASE
    """


class SPGreaseSchema(ProfileSchema):
    """Schema for GREASE
    """

    __profile_class__ = SPGrease
    version_tolerance = FieldsEnumString(enum_class=tls.ScanState)
    cipher_suite_tolerance = FieldsEnumString(enum_class=tls.ScanState)
    extension_tolerance = FieldsEnumString(enum_class=tls.ScanState)
    group_tolerance = FieldsEnumString(enum_class=tls.ScanState)
    sig_algo_tolerance = FieldsEnumString(enum_class=tls.ScanState)
    psk_mode_tolerance = FieldsEnumString(enum_class=tls.ScanState)


class SPEphemeralKeyReuse(SPObject):
    """Data class for ephemeral key reuse
    """


class SPEphemeralKeyReuseSchema(ProfileSchema):
    """Schema for ephemeral key reuse
    """

    __profile_class__ = SPEphemeralKeyReuse
    tls12_dhe_reuse = FieldsEnumString(enum_class=tls.ScanState)
    tls12_ecdhe_reuse = FieldsEnumString(enum_class=tls.ScanState)
    tls13_dhe_reuse = FieldsEnumString(enum_class=tls.ScanState)
    tls13_ecdhe_reuse = FieldsEnumString(enum_class=tls.ScanState)


class SPFeatures(SPObject):
    """Data class for TLS features.
    """


class SPFeaturesSchema(ProfileSchema):
    """Schema for TLS features.
    """

    __profile_class__ = SPFeatures
    compression = fields.List(fields.Nested(SPCompressionEnumSchema))
    encrypt_then_mac = FieldsEnumString(enum_class=tls.ScanState)
    extended_master_secret = FieldsEnumString(enum_class=tls.ScanState)
    session_id = FieldsEnumString(enum_class=tls.ScanState)
    session_ticket = FieldsEnumString(enum_class=tls.ScanState)
    session_ticket_lifetime = fields.Integer()
    resumption_psk = FieldsEnumString(enum_class=tls.ScanState)
    early_data = FieldsEnumString(enum_class=tls.ScanState)
    psk_lifetime = fields.Integer()
    insecure_renegotiation = FieldsEnumString(enum_class=tls.ScanState)
    secure_renegotation = FieldsEnumString(enum_class=tls.ScanState)
    scsv_renegotiation = FieldsEnumString(enum_class=tls.ScanState)
    heartbeat = FieldsEnumString(enum_class=tls.HeartbeatState)
    grease = fields.Nested(SPGreaseSchema)
    ephemeral_key_reuse = fields.Nested(SPEphemeralKeyReuseSchema)
    ocsp_stapling = FieldsEnumString(enum_class=tls.ScanState)
    ocsp_multi_stapling = FieldsEnumString(enum_class=tls.ScanState)
    downgrade_attack_prevention = FieldsEnumString(enum_class=tls.ScanState)


class SPPublicKey(SPObject):
    """Data class for all types of public key
    """

    def __init__(self, pub_key: Any = None, **kwargs) -> None:
        super().__init__(**kwargs)
        if not pub_key:
            return

        if isinstance(pub_key, rsa.RSAPublicKey):
            self.key_size = pub_key.key_size
            self.key_type = tls.SignatureAlgorithm.RSA
            pub_numbers = pub_key.public_numbers()
            self.key_exponent = pub_numbers.e
            self.key = pub_numbers.n.to_bytes(int(pub_key.key_size / 8), "big")

        elif isinstance(pub_key, dsa.DSAPublicKey):
            self.key_size = pub_key.key_size
            self.key_type = tls.SignatureAlgorithm.DSA
            pub_numbers = pub_key.public_numbers()  # type: ignore
            self.key = pub_numbers.y.to_bytes(  # type: ignore
                int(pub_key.key_size / 8), "big"
            )

            self.key_p = utils.int_to_bytes(
                pub_numbers.parameter_numbers.p  # type:ignore
            )
            self.key_q = utils.int_to_bytes(
                pub_numbers.parameter_numbers.q  # type:ignore
            )
            self.key_g = utils.int_to_bytes(
                pub_numbers.parameter_numbers.g  # type:ignore
            )

        elif isinstance(pub_key, ec.EllipticCurvePublicKey):
            self.key_size = pub_key.key_size
            self.key_type = tls.SignatureAlgorithm.ECDSA
            group = mappings.curve_to_group.get(pub_key.curve.name)
            if group is None:
                raise ValueError("curve {pub_key.curve.name} unknown")
            self.curve = group
            self.key = pub_key.public_bytes(
                Encoding.X962, PublicFormat.UncompressedPoint
            )

        elif isinstance(pub_key, ed25519.Ed25519PublicKey):
            self.key_size = 256
            self.key_type = tls.SignatureAlgorithm.ED25519
            self.key = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)

        elif isinstance(pub_key, ed448.Ed448PublicKey):
            self.key_size = 456
            self.key_type = tls.SignatureAlgorithm.ED448
            self.key = pub_key.public_bytes(Encoding.Raw, PublicFormat.Raw)


class SPPubKeyRsaSchema(ProfileSchema):
    """Schema for RSA public key.
    """

    __profile_class__ = SPPublicKey
    key_type = FieldsEnumString(enum_class=tls.SignatureAlgorithm)
    key = FieldsBytes()
    key_exponent = fields.Integer()
    key_size = fields.Integer()


class SPPubKeyDsaSchema(ProfileSchema):
    """Schema for DSA public key.
    """

    __profile_class__ = SPPublicKey
    key_type = FieldsEnumString(enum_class=tls.SignatureAlgorithm)
    key = FieldsBytes()
    key_p = FieldsBytes()
    key_q = FieldsBytes()
    key_g = FieldsBytes()


class SPPubKeyEcdsaSchema(ProfileSchema):
    """Schema for ECDSA public key.
    """

    __profile_class__ = SPPublicKey
    key_type = FieldsEnumString(enum_class=tls.SignatureAlgorithm)
    key = FieldsBytes()
    curve = FieldsEnumString(enum_class=tls.SupportedGroups)
    key_size = fields.Integer()


class SPPubKeyEdSchema(ProfileSchema):
    """Schema for ED25519 and ED448 public key
    """

    __profile_class__ = SPPublicKey
    key_type = FieldsEnumString(enum_class=tls.SignatureAlgorithm)
    key = FieldsBytes()


class SPPublicKeySchema(OneOfSchema):
    """Choice of public key types
    """

    type_field = "key_type"
    type_field_remove = False
    type_schemas = {
        "ECDSA": SPPubKeyEcdsaSchema,
        "RSA": SPPubKeyRsaSchema,
        "DSA": SPPubKeyDsaSchema,
        "ED25519": SPPubKeyEdSchema,
        "ED448": SPPubKeyEdSchema,
    }

    def get_obj_type(self, obj: Any) -> str:
        return obj.key_type.name


class SPCertGeneralName(SPObject):
    """Data class for general name
    """

    # TODO: replace Any by x509 class
    def __init__(self, name: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not name:
            return

        if isinstance(name, x509.RFC822Name):
            self.rfc822_name = name.value

        elif isinstance(name, x509.DNSName):
            self.dns_name = name.value

        elif isinstance(name, x509.DirectoryName):
            self.directory_name = name.value.rfc4514_string()

        elif isinstance(name, x509.UniformResourceIdentifier):
            self.uri = name.value

        elif isinstance(name, x509.IPAddress):
            self.ip_address = name.value

        else:
            logging.error(f"certificate extension: general name {name} not supported")


class SPCertGeneralNameSchema(ProfileSchema):
    """Schema for general name
    """

    __profile_class__ = SPCertGeneralName
    rfc822_name = fields.String()
    dns_name = fields.String()
    directory_name = fields.String()
    uri = fields.String()
    ip_address = fields.String()
    registered_id = fields.String()


# TODO: implement other name
#    other_name = fields.Nested(SpCertOtherNameSchema)


class SPCertNoticeReference(SPObject):
    """Data class for notice reference
    """

    # TODO: replace Any by x509 class
    def __init__(self, ref: Optional[Any] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        if not ref:
            return

        self.organization = ref.organization
        self.notice_numbers = [number for number in ref.notice_numbers]


class SPCertNoticeReferenceSchema(ProfileSchema):
    """Schema for notice reference
    """

    __profile_class__ = SPCertNoticeReference
    organization = fields.String()
    notice_numbers = fields.List(fields.Integer())


class SPCertUserNotice(SPObject):
    """Data class for user notice
    """

    # TODO: replace Any by x509 class
    def __init__(self, notice: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not notice:
            return

        if isinstance(notice, x509.UserNotice):
            self.explicit_text = notice.explicit_text
            if notice.notice_reference is not None:
                self.notice_reference = SPCertNoticeReference(
                    ref=notice.notice_reference
                )

        else:
            self.text = notice


class SPCertUserNoticeSchema(ProfileSchema):
    """Schema for user notice
    """

    __profile_class__ = SPCertUserNotice
    explicit_text = fields.String()
    notice_reference = fields.Nested(SPCertNoticeReferenceSchema)
    text = fields.String()


class SPDistrPoint(SPObject):
    """Data class for distribution point
    """

    # TODO: replace Any by x509 class
    def __init__(self, point: Optional[Any] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        if not point:
            return

        if point.full_name is not None:
            self.full_name = [SPCertGeneralName(name=name) for name in point.full_name]

        if point.relative_name is not None:
            logging.error(
                "Certificate extensions CrlDistrPoints: relative name "
                "not implemented"
            )

        if point.crl_issuer is not None:
            logging.error(
                "Certificate extensions CrlDistrPoints: crl_issuer not implemented"
            )

        if point.reasons is not None:
            logging.error(
                "Certificate extensions CrlDistrPoints: reasons not implemented"
            )


class SPDistrPointSchema(ProfileSchema):
    """Schema for distribution point
    """

    __profile_class__ = SPDistrPoint
    full_name = fields.List(fields.Nested(SPCertGeneralNameSchema))


class SPCertPolicyInfo(SPObject):
    """Data class for policy info
    """

    # TODO: replace Any by x509 class
    def __init__(self, policy: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not policy:
            return

        self.policy_name = policy.policy_identifier._name
        self.policy_oid = policy.policy_identifier.dotted_string
        if policy.policy_qualifiers is not None:
            self.policy_qualifiers = [
                SPCertUserNotice(notice=notice) for notice in policy.policy_qualifiers
            ]


class SPCertPolicyInfoSchema(ProfileSchema):
    """Schema for policy info
    """

    __profile_class__ = SPCertPolicyInfo
    policy_name = fields.String()
    policy_oid = fields.String()
    policy_qualifiers = fields.List(fields.Nested(SPCertUserNoticeSchema))


class SPCertAccessDescription(SPObject):
    """Data class for access description
    """

    # TODO: replace Any by x509 class
    def __init__(self, descr: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not descr:
            return

        self.access_method = descr.access_method._name
        self.access_location = descr.access_location.value


class SPCertAccessDescriptionSchema(ProfileSchema):
    """Schema for access description
    """

    __profile_class__ = SPCertAccessDescription
    access_method = fields.String()
    access_location = fields.String()


class SPCertSignedTimestamp(SPObject):
    """Data class for signed certificate timestamp
    """

    # TODO: replace Any by x509 class
    def __init__(self, timestamp: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not timestamp:
            return

        self.version = timestamp.version.name
        self.log_id = timestamp.log_id
        self.timestamp = timestamp.timestamp
        self.entry_type = timestamp.entry_type.name


class SPCertSignedTimestampSchema(ProfileSchema):
    """Schema for signed certificate timestamp
    """

    __profile_class__ = SPCertSignedTimestamp
    version = fields.String()
    log_id = FieldsBytes()
    timestamp = fields.DateTime()
    entry_type = fields.String()


class SPCertExtension(SPObject):
    """Data class for all extensions
    """

    def _ext_unrecognized_extension(self, value):
        self.bytes = value.value

    def _ext_key_usage(self, value):
        key_usage = []
        if value.digital_signature:
            key_usage.append(tls.CertKeyUsage.DIGITAL_SIGNATURE)

        if value.content_commitment:
            key_usage.append(tls.CertKeyUsage.CONTENT_COMMITMENT)

        if value.key_encipherment:
            key_usage.append(tls.CertKeyUsage.KEY_ENCIPHERMENT)

        if value.data_encipherment:
            key_usage.append(tls.CertKeyUsage.DATA_ENCIPHERMENT)

        if value.key_agreement:
            key_usage.append(tls.CertKeyUsage.KEY_AGREEMENT)
            if value.encipher_only:
                key_usage.append(tls.CertKeyUsage.ENCIPHER_ONLY)

            if value.decipher_only:
                key_usage.append(tls.CertKeyUsage.DECIPHER_ONLY)

        if value.key_cert_sign:
            key_usage.append(tls.CertKeyUsage.KEY_CERT_SIGN)

        if value.crl_sign:
            key_usage.append(tls.CertKeyUsage.CRL_SIGN)

        self.key_usage = key_usage

    def _ext_basic_constraints(self, value):
        self.ca = tls.ScanState(value.ca)
        if value.path_length is not None:
            self.path_length = value.path_length

    def _ext_extended_key_usage(self, value):
        self.extended_key_usage = [usage._name for usage in value]

    def _ext_subjec_key_id(self, value):
        self.digest = value.digest

    def _ext_ocsp_no_check(self, value):
        logging.error("Certificate extensions OcspNoCheck not implemented")

    def _ext_tls_features(self, value):
        self.tls_features = [tls.Extension(feature.value) for feature in value]

    def _ext_name_constraints(self, value):
        self.permitted_subtrees = [
            SPCertGeneralName(name=name) for name in value.permitted_subtrees
        ]
        self.excluded_subtrees = [
            SPCertGeneralName(name=name) for name in value.excluded_subtrees
        ]

    def _ext_authority_key_id(self, value):
        self.key_identifier = value.key_identifier
        if value.authority_cert_issuer is not None:
            self.authority_cert_issuer = [
                SPCertGeneralName(name=name) for name in value.authority_cert_issuer
            ]

        if value.authority_cert_serial_number is not None:
            self.authority_cert_serial_number = value.authority_cert_serial_number

    def _ext_subj_alt_name(self, value):
        self.subj_alt_names = [subj_alt_name.value for subj_alt_name in value]

    def _ext_issuer_alt_name(self, value):
        self.issuer_alt_name = [issuer_alt_name.value for issuer_alt_name in value]

    def _ext_precert_sign_cert_timestamps(self, value):
        self.signed_certificate_timestamps = [
            SPCertSignedTimestamp(timestamp=timestamp) for timestamp in value
        ]

    def _ext_precert_poison(self, value):
        logging.error("Certificate extensions PreCertPoison not implemented")

    def _ext_signed_cert_timestamps(self, value):
        logging.error("Certificate extensions SignedCertTimestamp not implemented")

    def _ext_delta_clr_indicator(self, value):
        logging.error("Certificate extensions DeltaClrIndicator not implemented")

    def _ext_authority_info_access(self, value):
        self.authority_info_access = [
            SPCertAccessDescription(descr=descr) for descr in value
        ]

    def _ext_subject_info_access(self, value):
        logging.error("Certificate extensions SubjectInfoAccess not implemented")

    def _ext_freshest_crl(self, value):
        logging.error("Certificate extensions FreshestCrl not implemented")

    def _ext_crl_distribution_points(self, value):
        self.distribution_points = [SPDistrPoint(point=point) for point in value]

    def _ext_inhibit_any_policy(self, value):
        logging.error("Certificate extensions InhibitAnyPolicy not implemented")

    def _ext_policy_constraints(self, value):
        if value.require_explicit_policy is not None:
            self.require_explicit_policy = value.require_explicit_policy

        if value.inhibit_policy_mapping is not None:
            self.inhibit_policy_mapping = value.inhibit_policy_mapping

    def _ext_crl_number(self, value):
        logging.error("Certificate extensions CrlNumber not implemented")

    def _ext_issuing_dist_point(self, value):
        logging.error("Certificate extensions IssuingDistPoint not implemented")

    def _ext_cert_policies(self, value):
        self.cert_policies = [SPCertPolicyInfo(policy=policy) for policy in value]

    def _ext_cert_issuer(self, value):
        raise NotImplementedError

    def _ext_crl_reason(self, value):
        logging.error("Certificate extensions CertIssuer not implemented")

    def _ext_invalidity_date(self, value):
        logging.error("Certificate extensions InvalidityDate not implemented")

    def _ext_ocsp_nonce(self, value):
        logging.error("Certificate extensions OcspNonce not implemented")

    _map_ext = {
        x509.extensions.KeyUsage: _ext_key_usage,
        x509.extensions.BasicConstraints: _ext_basic_constraints,
        x509.extensions.ExtendedKeyUsage: _ext_extended_key_usage,
        x509.extensions.OCSPNoCheck: _ext_ocsp_no_check,
        x509.extensions.TLSFeature: _ext_tls_features,
        x509.extensions.NameConstraints: _ext_name_constraints,
        x509.extensions.AuthorityKeyIdentifier: _ext_authority_key_id,
        x509.extensions.SubjectKeyIdentifier: _ext_subjec_key_id,
        x509.extensions.SubjectAlternativeName: _ext_subj_alt_name,
        x509.extensions.IssuerAlternativeName: _ext_issuer_alt_name,
        x509.extensions.PrecertificateSignedCertificateTimestamps: _ext_precert_sign_cert_timestamps,  # noqa: E501
        x509.extensions.PrecertPoison: _ext_precert_poison,
        x509.extensions.SignedCertificateTimestamps: _ext_signed_cert_timestamps,
        x509.extensions.DeltaCRLIndicator: _ext_delta_clr_indicator,
        x509.extensions.AuthorityInformationAccess: _ext_authority_info_access,
        x509.extensions.SubjectInformationAccess: _ext_subject_info_access,
        x509.extensions.FreshestCRL: _ext_freshest_crl,
        x509.extensions.CRLDistributionPoints: _ext_crl_distribution_points,
        x509.extensions.InhibitAnyPolicy: _ext_inhibit_any_policy,
        x509.extensions.PolicyConstraints: _ext_policy_constraints,
        x509.extensions.CRLNumber: _ext_crl_number,
        x509.extensions.IssuingDistributionPoint: _ext_issuing_dist_point,
        x509.extensions.UnrecognizedExtension: _ext_unrecognized_extension,
        x509.extensions.CertificatePolicies: _ext_cert_policies,
        x509.extensions.CertificateIssuer: _ext_cert_issuer,
        x509.extensions.CRLReason: _ext_crl_reason,
        x509.extensions.InvalidityDate: _ext_invalidity_date,
        x509.extensions.OCSPNonce: _ext_ocsp_nonce,
    }

    # TODO: replace Any by x509 class
    def __init__(self, ext: Optional[Any] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        if not ext:
            return

        self.name = ext.value.__class__.__name__
        self.oid = ext.oid.dotted_string
        self.criticality = tls.ScanState(ext.critical)
        # TODO: resolve type issue
        func = self._map_ext.get(type(ext.value))  # type: ignore
        if func is not None:
            func(self, ext.value)


class SPCertExtUnrecognizedExtensionSchema(ProfileSchema):
    """Data class for unknown certificate extension
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    bytes = FieldsBytes()


class SPCertExtKeyUsageSchema(ProfileSchema):
    """Data class for certificate extension KeyUsage
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    key_usage = fields.List(FieldsEnumString(enum_class=tls.CertKeyUsage))


class SPCertExtBasicConstraintsSchema(ProfileSchema):
    """Data class for certificate extension BasicConstraints
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    ca = FieldsEnumString(enum_class=tls.ScanState)
    path_length = fields.Integer()


class SPCertExtExtendedKeyUsageSchema(ProfileSchema):
    """Data class for certificate extension ExtendedKeyUsage
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    extended_key_usage = fields.List(fields.String())


class SPCertExtOCSPNoCheckSchema(ProfileSchema):
    """Data class for certificate extension OCSPNoCheck
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtTLSFeatureSchema(ProfileSchema):
    """Data class for certificate extension TLSFeature
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    tls_features = fields.List(FieldsEnumString(enum_class=tls.Extension))


class SPCertExtNameConstraintsSchema(ProfileSchema):
    """Data class for certificate extension NameConstraints
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    permitted_subtrees = fields.List(fields.Nested(SPCertGeneralNameSchema))
    excluded_subtrees = fields.List(fields.Nested(SPCertGeneralNameSchema))


class SPCertExtAuthorityKeyIdentifierSchema(ProfileSchema):
    """Data class for certificate extension AuthorityKeyIdentifier
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    key_identifier = FieldsBytes()
    authority_cert_issuer = fields.List(fields.Nested(SPCertGeneralNameSchema))
    authority_cert_serial_number = fields.Integer()


class SPCertExtSubjectKeyIdentifierSchema(ProfileSchema):
    """Data class for certificate extension SubjectKeyIdentifier
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    digest = FieldsBytes()


class SPCertExtSubjectAlternativeNameSchema(ProfileSchema):
    """Data class for certificate extension SubjectAlternativeName
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    subj_alt_names = fields.List(fields.String())


class SPCertExtIssuerAlternativeNameSchema(ProfileSchema):
    """Data class for certificate extension IssuerAlternativeName
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    issuer_alt_name = fields.List(fields.String())


class SPCertExtPrecertSignedCertTimestampsSchema(ProfileSchema):
    """Data class for certificate extension ExtPrecertSignedCertTimestamps
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    signed_certificate_timestamps = fields.List(
        fields.Nested(SPCertSignedTimestampSchema)
    )


class SPCertExtPrecertPoisonSchema(ProfileSchema):
    """Data class for certificate extension PrecertPoison
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtSignedCertificateTimestampsSchema(ProfileSchema):
    """Data class for certificate extension SignedCertificateTimestamps
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    signed_certificate_timestamps = fields.List(
        fields.Nested(SPCertSignedTimestampSchema)
    )


class SPCertExtDeltaCRLIndicatorSchema(ProfileSchema):
    """Data class for certificate extension DeltaCRLIndicator
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtAuthorityInformationAccessSchema(ProfileSchema):
    """Data class for certificate extension AuthorityInformationAccess
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    authority_info_access = fields.List(fields.Nested(SPCertAccessDescriptionSchema))


class SPCertExtSubjectInformationAccessSchema(ProfileSchema):
    """Data class for certificate extension SubjectInformationAccess
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtFreshestCRLSchema(ProfileSchema):
    """Data class for certificate extension FreshestCRL
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtCRLDistributionPointsSchema(ProfileSchema):
    """Data class for certificate extension CRLDistributionPoints
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    distribution_points = fields.List(fields.Nested(SPDistrPointSchema))


class SPCertExtInhibitAnyPolicySchema(ProfileSchema):
    """Data class for certificate extension InhibitAnyPolicy
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtPolicyConstraintsSchema(ProfileSchema):
    """Data class for certificate extension PolicyConstraints
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    require_explicit_policy = fields.Integer()
    inhibit_policy_mapping = fields.Integer()


class SPCertExtCRLNumberSchema(ProfileSchema):
    """Data class for certificate extension CRLNumber
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtIssuingDistributionPointSchema(ProfileSchema):
    """Data class for certificate extension IssuingDistributionPoint
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtCertificatePoliciesSchema(ProfileSchema):
    """Data class for certificate extension CertificatePolicies
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)
    cert_policies = fields.List(fields.Nested(SPCertPolicyInfoSchema))


class SPCertExtCertificateIssuerSchema(ProfileSchema):
    """Data class for certificate extension CertificateIssuer
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtCRLReasonSchema(ProfileSchema):
    """Data class for certificate extension CRLReason
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtInvalidityDateSchema(ProfileSchema):
    """Data class for certificate extension InvalidityDate
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtOCSPNonceSchema(ProfileSchema):
    """Data class for certificate extension OCSPNonce
    """

    __profile_class__ = SPCertExtension
    name = fields.String()
    oid = fields.String()
    criticality = FieldsEnumString(enum_class=tls.ScanState)


class SPCertExtensionSchema(OneOfSchema):
    """Choice of certificate extensions
    """

    type_field = "name"
    type_field_remove = False
    type_schemas = {
        "KeyUsage": SPCertExtKeyUsageSchema,
        "UnrecognizedExtension": SPCertExtUnrecognizedExtensionSchema,
        "BasicConstraints": SPCertExtBasicConstraintsSchema,
        "ExtendedKeyUsage": SPCertExtExtendedKeyUsageSchema,
        "OCSPNoCheck": SPCertExtOCSPNoCheckSchema,
        "TLSFeature": SPCertExtTLSFeatureSchema,
        "NameConstraints": SPCertExtNameConstraintsSchema,
        "AuthorityKeyIdentifier": SPCertExtAuthorityKeyIdentifierSchema,
        "SubjectKeyIdentifier": SPCertExtSubjectKeyIdentifierSchema,
        "SubjectAlternativeName": SPCertExtSubjectAlternativeNameSchema,
        "IssuerAlternativeName": SPCertExtIssuerAlternativeNameSchema,
        "PrecertificateSignedCertificateTimestamps": SPCertExtPrecertSignedCertTimestampsSchema,  # noqa: E501
        "PrecertPoison": SPCertExtPrecertPoisonSchema,
        "SignedCertificateTimestamps": SPCertExtSignedCertificateTimestampsSchema,
        "DeltaCRLIndicator": SPCertExtDeltaCRLIndicatorSchema,
        "AuthorityInformationAccess": SPCertExtAuthorityInformationAccessSchema,
        "SubjectInformationAccess": SPCertExtSubjectInformationAccessSchema,
        "FreshestCRL": SPCertExtFreshestCRLSchema,
        "CRLDistributionPoints": SPCertExtCRLDistributionPointsSchema,
        "InhibitAnyPolicy": SPCertExtInhibitAnyPolicySchema,
        "PolicyConstraints": SPCertExtPolicyConstraintsSchema,
        "CRLNumber": SPCertExtCRLNumberSchema,
        "IssuingDistributionPoint": SPCertExtIssuingDistributionPointSchema,
        "CertificatePolicies": SPCertExtCertificatePoliciesSchema,
        "CertificateIssuer": SPCertExtCertificateIssuerSchema,
        "CRLReason": SPCertExtCRLReasonSchema,
        "InvalidityDate": SPCertExtInvalidityDateSchema,
        "OCSPNonce": SPCertExtOCSPNonceSchema,
    }

    # TODO: add type annotations
    def get_obj_type(self, obj):
        return obj.name


class SPCertificate(SPObject):
    """Data class for a certificate.
    """

    def __init__(self, cert: Optional[crt.Certificate] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not cert:
            return

        self.pem = cert.pem.decode()
        self.subject = cert.subject_str
        self.issuer = cert.parsed.issuer.rfc4514_string()
        self.version = cert.parsed.version
        self.serial_number_int = cert.parsed.serial_number
        self.serial_number_bytes = utils.int_to_bytes(cert.parsed.serial_number)
        self.signature = cert.parsed.signature
        self.not_valid_before = cert.parsed.not_valid_before
        self.not_valid_after = cert.parsed.not_valid_after
        diff = cert.parsed.not_valid_after - cert.parsed.not_valid_before
        self.validity_period_days = diff.days
        self.self_signed = tls.ScanState(cert.self_signed)
        if cert.subject_matches is not None:
            self.subject_matches = tls.ScanState(cert.subject_matches)
        self.fingerprint_sha1 = cert.fingerprint_sha1
        self.fingerprint_sha256 = cert.fingerprint_sha256
        self.signature_algorithm = cert.signature_algorithm
        self.public_key = SPPublicKey(pub_key=cert.parsed.public_key())
        self.extensions = [SPCertExtension(ext=ext) for ext in cert.parsed.extensions]
        if cert.crl_status is not None:
            self.crl_revocation_status = cert.crl_status
        if cert.ocsp_status is not None:
            self.ocsp_revocation_status = cert.ocsp_status
        if cert.issues:
            self.issues = cert.issues
        self.ocsp_must_staple = cert.ocsp_must_staple
        self.ocsp_must_staple_multi = cert.ocsp_must_staple_multi
        self.extended_validation = cert.extended_validation
        self.from_trust_store = tls.ScanState(cert.from_trust_store)


class SPCertificateSchema(ProfileSchema):
    """Schema for a certificate.
    """

    __profile_class__ = SPCertificate
    pem = FieldsBlockString()
    subject = fields.String()
    issuer = fields.String()
    version = fields.String()
    serial_number_int = fields.Integer()
    serial_number_bytes = FieldsBytes()
    signature = FieldsBytes()
    not_valid_before = fields.DateTime()
    not_valid_after = fields.DateTime()
    validity_period_days = fields.Integer()
    self_signed = FieldsEnumString(enum_class=tls.ScanState)
    subject_matches = FieldsEnumString(enum_class=tls.ScanState)
    fingerprint_sha1 = FieldsBytes()
    fingerprint_sha256 = FieldsBytes()
    signature_algorithm = FieldsEnumString(enum_class=tls.SignatureScheme)
    public_key = fields.Nested(SPPublicKeySchema)
    crl_revocation_status = FieldsEnumString(enum_class=tls.CertCrlStatus)
    ocsp_revocation_status = FieldsEnumString(enum_class=tls.OcspStatus)
    ocsp_must_staple = FieldsEnumString(enum_class=tls.ScanState)
    ocsp_must_staple_multi = FieldsEnumString(enum_class=tls.ScanState)
    extensions = fields.List(fields.Nested(SPCertExtensionSchema))
    extended_validation = FieldsEnumString(enum_class=tls.ScanState)
    issues = fields.List(fields.String())
    from_trust_store = FieldsEnumString(enum_class=tls.ScanState)


class SPCertChain(SPObject):
    """Data class for a certificate chain.
    """

    def __init__(
        self, chain: Optional["cert_chain.CertChain"] = None, **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)
        if not chain:
            return

        # TODO: resolve type issue
        self.id = chain.id  # type: ignore
        self.successful_validation = tls.ScanState(chain.successful_validation)
        self.cert_chain = [SPCertificate(cert=cert) for cert in chain.certificates]
        if chain.issues:
            self.issues = chain.issues

        self.root_cert_transmitted = tls.ScanState(chain.root_cert_transmitted)
        if chain.root_cert is not None:
            self.root_certificate = SPCertificate(cert=chain.root_cert)


class SPCertChainSchema(ProfileSchema):
    """Schema for a certificate chain.
    """

    __profile_class__ = SPCertChain
    id = fields.Integer()
    successful_validation = FieldsEnumString(enum_class=tls.ScanState)
    cert_chain = fields.List(fields.Nested(SPCertificateSchema))
    issues = fields.List(fields.String())
    root_cert_transmitted = FieldsEnumString(enum_class=tls.ScanState)
    root_certificate = fields.Nested(SPCertificateSchema)


class SPVersionEnumSchema(ProfileEnumSchema):
    """Schema for a TLS version (enum).
    """

    __profile_class__ = tls.Version


class SPSupportedGroupEnumSchema(ProfileEnumSchema):
    """Schema for a supported group (enum).
    """

    __profile_class__ = tls.SupportedGroups


class SPSupportedGroups(SPObject):
    """Data class for supported groups.
    """


class SPSupportedGroupsSchema(ProfileSchema):
    """Data class for supported groups.
    """

    __profile_class__ = SPSupportedGroups
    extension_supported = FieldsEnumString(enum_class=tls.ScanState)
    server_preference = FieldsEnumString(enum_class=tls.ScanState)
    groups_advertised = FieldsEnumString(enum_class=tls.ScanState)
    groups = fields.List(fields.Nested(SPSupportedGroupEnumSchema))


class SPSigAlgoEnumSchema(ProfileEnumSchema):
    """Schema for signature algorithms (enum).
    """

    __profile_class__ = tls.SignatureScheme


class SPSignatureAlgorithms(SPObject):
    """Data class for signature algorithms
    """

    def __init__(self, **kwargs: Any) -> None:
        # TODO: provide type of list
        self.algorithms = []  # type: ignore
        super().__init__(**kwargs)


class SPSignatureAlgorithmsSchema(ProfileSchema):
    """Schema for signature algorithms
    """

    __profile_class__ = SPSignatureAlgorithms
    algorithms = fields.List(fields.Nested(SPSigAlgoEnumSchema))
    info = fields.List(fields.String)


class SPCipherSuiteSchema(ProfileEnumSchema):
    """Schema for a cipher suite (enum).
    """

    __profile_class__ = tls.CipherSuite


class SPRecordProtocolSchema(ProfileEnumSchema):
    """Schema for a record protocol type.
    """

    __profile_class__ = tls.ContentType


class SPCiphers(SPObject):
    """Data class for ciphers
    """


class SPCiphersSchema(ProfileSchema):
    """Schema for ciphers
    """

    __profile_class__ = SPCiphers
    cipher_suites = fields.List(fields.Nested(SPCipherSuiteSchema))
    server_preference = FieldsEnumString(enum_class=tls.ScanState)
    chacha_poly_preference = FieldsEnumString(enum_class=tls.ScanState)


class SPCipherKindSchema(ProfileEnumSchema):
    """Schema for an SSL2 cipher kind (enum).
    """

    __profile_class__ = tls.SSLCipherKind


class SPDhGroup(SPObject):
    """Data class for DH groups
    """


class SPDhGroupSchema(ProfileSchema):
    """Schema for DH groups
    """

    __profile_class__ = SPDhGroup
    name = fields.String()
    size = fields.Integer()
    g_value = fields.Integer()
    p_value = FieldsBytes()


class SPNameResolution(SPObject):
    """Data class for name resolution infos
    """


class SPNameResolutionSchema(ProfileSchema):
    """Schema for name resolution infos
    """

    __profile_class__ = SPNameResolution
    domain_name = fields.String()
    ipv4_addresses = fields.List(fields.String())
    ipv6_addresses = fields.List(fields.String())


class SPServer(SPObject):
    """Data class for the servers' information
    """


class SPServerSchema(ProfileSchema):
    """Schema for the server's information
    """

    __profile_class__ = SPServer
    ip = fields.String()
    name_resolution = fields.Nested(SPNameResolutionSchema)
    port = fields.Integer()
    sni = fields.String()


class SPVersion(SPObject):
    """Data class for a dedicated TLS version.
    """


class SPVersionSchema(ProfileSchema):
    """Schema for a dedicated TLS version.
    """

    __profile_class__ = SPVersion
    cipher_kinds = fields.List(fields.Nested(SPCipherKindSchema))
    ciphers = fields.Nested(SPCiphersSchema)
    dh_group = fields.Nested(SPDhGroupSchema)
    supported_groups = fields.Nested(SPSupportedGroupsSchema)
    signature_algorithms = fields.Nested(SPSignatureAlgorithmsSchema)
    version = fields.Nested(SPVersionEnumSchema)
    support = FieldsEnumString(enum_class=tls.ScanState)


class SPCipherGroup(SPObject):
    """Data class for a cipher group (cipher suite, tls version, record protocol)
    """


class SPCipherGroupSchema(ProfileSchema):
    """Schema for a cipher group (cipher suite, TLS version, record protocol)
    """

    __profile_class__ = SPCipherGroup
    version = fields.Nested(SPVersionEnumSchema)
    cipher_suite = fields.Nested(SPCipherSuiteSchema)
    record_protocol = fields.Nested(SPRecordProtocolSchema)


class SPCbcPaddingOracle(SPObject):
    """Data class for CBC padding oracles
    """


class SPCbcPaddingOracleSchema(ProfileSchema):
    """Schema for CBC padding oracles
    """

    __profile_class__ = SPCbcPaddingOracle
    observable = FieldsEnumString(enum_class=tls.ScanState)
    strong = FieldsEnumString(enum_class=tls.ScanState)
    types = fields.List(FieldsEnumString(enum_class=tls.SPCbcPaddingOracle))
    cipher_group = fields.List(fields.Nested(SPCipherGroupSchema))


class SPCbcPaddingOracleInfo(SPObject):
    """Data class for CBC padding oracle info
    """


class SPCbcPaddingOracleInfoSchema(ProfileSchema):
    """Schema for CBC padding oracle info
    """

    __profile_class__ = SPCbcPaddingOracleInfo
    vulnerable = FieldsEnumString(enum_class=tls.ScanState)
    accuracy = FieldsEnumString(enum_class=tls.OracleScanAccuracy)
    oracles = fields.List(fields.Nested(SPCbcPaddingOracleSchema))


class SPVulnerabilities(SPObject):
    """Data class for vulnerabilities
    """


class SPVulnerabilitiesSchema(ProfileSchema):
    """Schema for vulnerabilities
    """

    __profile_class__ = SPVulnerabilities
    ccs_injection = FieldsEnumString(enum_class=tls.ScanState)
    heartbleed = FieldsEnumString(enum_class=tls.HeartbleedStatus)
    robot = FieldsEnumString(enum_class=tls.RobotVulnerability)
    poodle = FieldsEnumString(enum_class=tls.ScanState)
    tls_poodle = FieldsEnumString(enum_class=tls.ScanState)
    lucky_minus_20 = FieldsEnumString(enum_class=tls.ScanState)
    cbc_padding_oracle = fields.Nested(SPCbcPaddingOracleInfoSchema)
    beast = FieldsEnumString(enum_class=tls.ScanState)
    crime = FieldsEnumString(enum_class=tls.ScanState)
    sweet_32 = FieldsEnumString(enum_class=tls.ScanState)
    freak = FieldsEnumString(enum_class=tls.ScanState)
    logjam = FieldsEnumString(enum_class=tls.Logjam)


class SPMalfunctionIssue(SPObject):
    """Data class for a server issue
    """

    # TODO: replace Any
    def __init__(self, issue: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not issue:
            return

        self.name = issue.name
        self.description = issue.value


class SPMalfunctionIssueSchema(ProfileSchema):
    """Schema for malfunction issue
    """

    __profile_class__ = SPMalfunctionIssue
    name = fields.String()
    description = fields.String()


class SPMalfunctionMessageSchema(ProfileEnumSchema):
    """Schema for malfunction message
    """

    __profile_class__ = tls.HandshakeType


class SPMalfunctionExtensionSchema(ProfileEnumSchema):
    """Schema for malfunction extension
    """

    __profile_class__ = tls.Extension


class SPServerMalfunction(SPObject):
    """Data class for server malfunction
    """

    # TODO: replace Any
    def __init__(self, malfunction: Optional[Any] = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if not malfunction:
            return

        self.issue = SPMalfunctionIssue(issue=malfunction.issue)
        if malfunction.message:
            self.message = malfunction.message

        if malfunction.extension:
            self.extension = malfunction.extension


class SPServerMalfunctionSchema(ProfileSchema):
    """Schema for server malfunction
    """

    __profile_class__ = SPServerMalfunction
    issue = fields.Nested(SPMalfunctionIssueSchema)
    message = fields.Nested(SPMalfunctionMessageSchema)
    extension = fields.Nested(SPMalfunctionExtensionSchema)


class ServerProfile(SPObject):
    """Data class for the server profile.

    Attributes:
        cert_chains (list :obj:`SPCertChain`): the list of certificate chains used
            by the server
        features (:obj:`SPFeatures`): the profile structure for the features supported
            by the server
        scan_info (:obj:`SPScanInfo`): object describing basic scan information
        server (:obj:`SPServer`): object describing the server's details
        versions (list of :obj:`SPVersion`): list versions supported by the server
        vulnerabilities (:obj:`SPVulnerabilities`): object containing infos regarding
            the vulnerabilities
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self._hash: Dict[int, int] = {}

    def allocate_versions(self) -> None:
        """Ensure that the versions property and cert_chains properties are setup.
        """

        if not hasattr(self, "versions"):
            self.versions: List[SPVersion] = []

        if not hasattr(self, "cert_chains"):
            self.cert_chains: List[SPCertChain] = []

    def allocate_features(self) -> None:
        """Ensure that the features property is setup.
        """

        if not hasattr(self, "features"):
            self.features = SPFeatures()

    def allocate_vulnerabilities(self) -> None:
        """Ensure that the vulnerabilities property is setup.
        """

        if not hasattr(self, "vulnerabilities"):
            self.vulnerabilities = SPVulnerabilities()

    def append_unique_cert_chain(self, chain: "cert_chain.CertChain") -> None:
        """Append a certificate chain to the profile, if not yet present.

        Arguments:
            chain (:obj:`tlsmate.cert.CertChain`): the chain to be added

        Returns:
            int: the index of the chain, which may be created newly, or it might have
            been present already.
        """

        if chain.digest in self._hash:
            return self._hash[chain.digest]  # type: ignore

        idx = len(self._hash) + 1
        self._hash[chain.digest] = idx
        # TODO: resolve type issue
        chain.id = idx  # type: ignore
        self.cert_chains.append(SPCertChain(chain=chain))

    def get_versions(
        self, exclude: Optional[List[tls.Version]] = None
    ) -> List[tls.Version]:
        """Get all TLS versions from the profile.

        Returns:
            list(:obj:`tlsmate.tls.Version`): A list of all supported versions.
        """

        if not hasattr(self, "versions"):
            return []

        if exclude is None:
            exclude = []

        return [
            obj.version  # type: ignore
            for obj in self.versions
            if (
                obj.version not in exclude  # type: ignore
                and obj.support is tls.ScanState.TRUE  # type: ignore
            )
        ]

    def get_version_profile(self, version: tls.Version) -> Optional[SPVersion]:
        """Get the profile entry for a given version.

        Returns:
            :obj:`SPVersion`: the profile entry for the given version or None, if the
            version is not supported by the server.
        """

        for vers_obj in self.versions:
            if (
                vers_obj.version is version  # type: ignore
                and vers_obj.support is tls.ScanState.TRUE  # type: ignore
            ):
                return vers_obj

        return None

    def get_cipher_suites(
        self, version: tls.Version
    ) -> Optional[List[tls.CipherSuite]]:
        """Get all cipher suites for a given TLS version.

        Arguments:
            version (:obj:`tlsmate.tls.Version`): The version for which to get
                the cipher suites for.

        Returns:
            list(:obj:`tlsmate.tls.Ciphersuite`): The list of cipher suites or None,
            if the version is not supported by the server.
        """

        version_prof = self.get_version_profile(version)
        if version_prof is not None:
            if version is tls.Version.SSL20:
                return version_prof.cipher_kinds  # type: ignore

            else:
                return version_prof.ciphers.cipher_suites  # type: ignore

        return None

    def get_supported_groups(
        self, version: tls.Version
    ) -> Optional[List[tls.SupportedGroups]]:
        """Get all supported groups for a given TLS version.

        Arguments:
            version (:class:`tlsmate.tls.Version`): the TLS version to use

        Returns:
            list: a list of all supported groups supported by the server for the given
            TLS version, or None if no supported groups are available.
        """

        version_prof = self.get_version_profile(version)
        try:
            return version_prof.supported_groups.groups  # type: ignore

        except AttributeError:
            return None

    def get_signature_algorithms(
        self, version: tls.Version
    ) -> Optional[List[tls.SignatureScheme]]:
        """Get all signature algorithms for a given TLS version.

        Arguments:
            version (:class:`tlsmate.tls.Version`): the TLS version to use

        Returns:
            list: a list of all signature algorithms supported by the server for the
            given TLS version, or None if no signature algorithms are available.
        """

        version_prof = self.get_version_profile(version)
        if version_prof is not None and hasattr(version_prof, "signature_algorithms"):
            return version_prof.signature_algorithms.algorithms  # type: ignore

        return None

    def get_cert_sig_algos(
        self, key_types: Optional[List[tls.SignatureAlgorithm]] = None
    ) -> List[tls.SignatureScheme]:
        """Get all signature algorithms from the cert chains for the given key_types

        Arguments:
            key_types (list (:obj:`tlsmate.tls.SignatureAlgorithm`): the type of the
                public key required in the host certificate. Can be None to get
                all signature algorithms of all certs of all chains.

        Returns:
            list(:obj:`tlsmate.tls.SignatureScheme`): the list requested
        """
        sig_algos = []
        for chain in self.cert_chains:
            key_type = chain.cert_chain[0].public_key.key_type
            if key_types is None or key_type in key_types:
                for cert in chain.cert_chain:
                    if cert.signature_algorithm not in sig_algos:
                        sig_algos.append(cert.signature_algorithm)

                if hasattr(chain, "root_certificate"):
                    if chain.root_certificate.signature_algorithm not in sig_algos:
                        sig_algos.append(chain.root_certificate.signature_algorithm)

        return sig_algos

    def get_profile_values(
        self, filter_versions: List[tls.Version], full_hs: bool = False
    ) -> structs.ProfileValues:
        """Get a set of some common attributes for the given TLS version(s).

        Arguments:
            filter_versions (list of :class:`tlsmate.tls.Version`): the list of
                TLS versions to retrieve the data for
            full_hs (bool): an indication if only those cipher suites shall be returned
                for which a full handshake is supported. Defaults to False.

        Returns:
            :obj:`tlsmate.structs.ProfileValues`: a structure that provides a list of
            the versions, the cipher suites, the supported groups and the
            signature algorithms
        """

        versions: List[tls.Version] = []
        cipher_suites: List[tls.CipherSuite] = []
        sig_algos: List[tls.SignatureScheme] = []
        groups: List[tls.SupportedGroups] = []
        key_shares: List[tls.SupportedGroups] = []

        # We want to treat higer protocol versions first, so that the result
        # provides the most desirable preference
        for version in sorted(self.get_versions(), reverse=True):
            if version not in filter_versions:
                continue

            # So the versions are restored in order from low to high
            versions.insert(0, version)

            vers_cs = self.get_cipher_suites(version)
            cipher_suites.extend(
                [cs for cs in vers_cs if cs not in cipher_suites]  # type: ignore
            )

            vers_sig = self.get_signature_algorithms(version)
            if vers_sig is not None:
                sig_algos.extend([algo for algo in vers_sig if algo not in sig_algos])

            # Add the signature algorithms used in the certificate chains as well, if
            # not yet present.
            sig_algos.extend(
                [
                    algo
                    for algo in self.get_cert_sig_algos()  # type: ignore
                    if algo not in sig_algos
                ]
            )
            vers_group = self.get_supported_groups(version)
            if vers_group is not None:
                groups.extend([group for group in vers_group if group not in groups])

            if version is tls.Version.TLS13:
                key_shares = vers_group  # type: ignore

        if full_hs:
            cipher_suites = utils.filter_cipher_suites(cipher_suites, full_hs=True)

        return structs.ProfileValues(
            versions=versions,
            cipher_suites=cipher_suites,
            supported_groups=groups,
            signature_algorithms=sig_algos,
            key_shares=key_shares,
        )

    def make_serializable(self) -> Dict[str, Any]:
        """Convert the object into seralizable types

        Returns:
            dict: the serializable data provided as a dict.
        """

        return ServerProfileSchema().dump(self)

    def load(self, data: Any) -> None:
        ServerProfileSchema(profile=self).load(data)


class ServerProfileSchema(ProfileSchema):
    """Base schema for the server profile.
    """

    __profile_class__ = ServerProfile
    cert_chains = fields.List(fields.Nested(SPCertChainSchema))
    features = fields.Nested(SPFeaturesSchema)
    scan_info = fields.Nested(SPScanInfoSchema)
    server = fields.Nested(SPServerSchema)
    server_malfunctions = fields.List(fields.Nested(SPServerMalfunctionSchema))
    versions = fields.List(fields.Nested(SPVersionSchema))
    vulnerabilities = fields.Nested(SPVulnerabilitiesSchema)

    def __init__(self, profile: Optional[ServerProfile] = None, **kwargs: Any) -> None:
        self._profile = profile
        super().__init__(**kwargs)

    @post_load
    def deserialize(self, data: Any, **kwargs: Any) -> None:
        super().deserialize(data, reuse_object=self._profile, **kwargs)



# -*- coding: utf-8 -*-
"""Module for handling a certificate chain
"""
# import basic stuff
import logging
import time
import datetime
from typing import List, Optional

# import own stuff
import tlsmate.cert as crt
import tlsmate.cert_utils as cert_utils
import tlsmate.recorder as recorder
import tlsmate.tls as tls

# import other stuff
import requests
from cryptography import x509
from cryptography.x509 import ocsp
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.x509.oid import AuthorityInformationAccessOID
from cryptography.exceptions import InvalidSignature


class CertChain(object):
    """Class representing a certificate chain.

    This object is iterable, yielding the x509 representation of the certificates.
    """

    def __init__(self):
        from tlsmate.tlsmate import TlsMate

        tlsmate = TlsMate.instance
        self._cert_chain_cache = tlsmate.cert_chain_cache
        self.certificates = []
        self._digest = hashes.Hash(hashes.SHA256())
        self._digest_value = None
        self._raise_on_failure = True
        self.issues = []
        self.successful_validation = False
        self.root_cert = None
        self.root_cert_transmitted = False
        self._recorder = tlsmate.recorder
        self._config = tlsmate.config
        self._trust_store = tlsmate.trust_store
        self._crl_manager = tlsmate.crl_manager
        self._trust_path = None
        self._proxies = None
        proxy = self._config.get("proxy")
        if proxy:
            self._proxies = dict(http=proxy, https=proxy)

    def append_bin_cert(self, bin_cert: bytes) -> None:
        """Append the chain by a certificate given in raw format.

        Arguments:
            bin_cert: the certificate to append in raw format
        """
        self.certificates.append(crt.Certificate(der=bin_cert, parse=True))
        self._digest.update(bin_cert)

    def append_pem_cert(self, pem_cert: bytes) -> None:
        """Append the chain by a certificate given in pem format.

        Arguments:
            pem_cert: the certificate to append in pem format
        """
        cert = crt.Certificate(pem=pem_cert)
        self.certificates.append(cert)
        self._digest.update(cert.bytes)

    @property
    def digest(self):
        """bytes: a SHA256 digest of the complete chain, usable for comparison"""
        if self._digest_value is None:
            self._digest_value = self._digest.finalize()
        return self._digest_value

    def _valid_crl(self, cert, issuer_cert, timestamp):
        """Check the CRL state for the given certificate."""

        if not self._config.get("crl"):
            cert.crl_status = tls.CertCrlStatus.UNDETERMINED
            return True

        try:
            dist_points = cert.parsed.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.CRL_DISTRIBUTION_POINTS
            )
        except x509.ExtensionNotFound:
            return True

        crl_urls = []
        for dist_point in dist_points.value:
            if dist_point.full_name is not None:
                for gen_name in dist_point.full_name:
                    if isinstance(gen_name, x509.UniformResourceIdentifier):
                        if gen_name.value.startswith("http://"):
                            crl_urls.append(gen_name.value)
            elif dist_point.relative_me is not None:
                raise NotImplementedError

        cert.crl_status = self._crl_manager.get_crl_status(
            crl_urls,
            cert.parsed.serial_number,
            cert.parsed.issuer,
            issuer_cert,
            timestamp,
            self._proxies,
        )
        logging.debug(f'CRL status is {cert.crl_status} for certificate "{cert}"')
        if cert.crl_status is not tls.CertCrlStatus.NOT_REVOKED:
            cert.mark_untrusted(f"CRL status not ok: {cert.crl_status}")
            return False

        return True

    def _verify_ocsp_resp(self, cert, response, timestamp, issuer_cert):
        if not response:
            return None

        ocsp_decoded = ocsp.load_der_ocsp_response(response)

        if ocsp_decoded.response_status is not ocsp.OCSPResponseStatus.SUCCESSFUL:
            return tls.OcspStatus.INVALID_RESPONSE

        if ocsp_decoded.certificates:
            sig_cert = crt.Certificate(
                x509_cert=ocsp_decoded.certificates[0], parse=True
            )
            self._determine_trust_path(sig_cert, -1, timestamp, None, False)
            if sig_cert.trusted is tls.ScanState.FALSE:
                return tls.OcspStatus.INVALID_ISSUER_CERT

        else:
            sig_cert = issuer_cert

        if sig_cert is None:
            return tls.OcspStatus.NO_ISSUER

        else:
            # check signature
            try:
                sig_scheme = cert_utils.map_x509_sig_scheme(
                    ocsp_decoded.signature_hash_algorithm,
                    ocsp_decoded.signature_algorithm_oid,
                )
                sig_cert.validate_signature(
                    sig_scheme, ocsp_decoded.tbs_response_bytes, ocsp_decoded.signature,
                )

            except InvalidSignature:
                return tls.OcspStatus.SIGNATURE_INVALID

        if ocsp_decoded.this_update > timestamp:
            return tls.OcspStatus.INVALID_TIMESTAMP

        if ocsp_decoded.next_update and ocsp_decoded.next_update < timestamp:
            return tls.OcspStatus.INVALID_TIMESTAMP

        if ocsp_decoded.certificate_status == ocsp.OCSPCertStatus.GOOD:
            return tls.OcspStatus.NOT_REVOKED

        elif ocsp_decoded.certificate_status == ocsp.OCSPCertStatus.REVOKED:
            return tls.OcspStatus.REVOKED

        else:
            return tls.OcspStatus.UNKNOWN

    def verify_ocsp_stapling(
        self, responses: List[bytes], raise_on_failure: bool
    ) -> List[Optional[tls.OcspStatus]]:
        """Check the status for a OCSP response

        Arguments:
            responses: the OCSP response
            raise_on_failure: An indication, if the validation shall abort
                in case exceptional cases are detected. Normally set to False for
                server scans.

        Returns:
            the status, one for each response. An list element can be None, if
            an empty response is provided.
        """

        ret_status: List[Optional[tls.OcspStatus]] = []
        for idx, resp in enumerate(responses):
            if not resp:
                ret_status.append(None)
                continue

            cert = self.certificates[idx]
            if not self.successful_validation:
                ocsp_status = tls.OcspStatus.INVALID_ISSUER_CERT

            else:
                ocsp_status = self._verify_ocsp_resp(
                    cert, resp, self._recorder.get_timestamp(), cert.issuer_cert,
                )
            ret_status.append(ocsp_status)
            issue = f"OCSP stapling status {ocsp_status} for certificate {cert}"
            logging.debug(issue)
            if ocsp_status is not tls.OcspStatus.NOT_REVOKED and raise_on_failure:
                raise tls.UntrustedCertificate(issue)

        return ret_status

    def _valid_ocsp(self, cert, issuer_cert, timestamp):
        """Check the OCSP status for the given certificate."""

        if not self._config.get("ocsp"):
            cert.ocsp_status = tls.OcspStatus.UNDETERMINED
            return True

        try:
            aia = cert.parsed.extensions.get_extension_for_oid(
                x509.oid.ExtensionOID.AUTHORITY_INFORMATION_ACCESS
            ).value

        except x509.ExtensionNotFound:
            return True

        ocsps = [
            ia for ia in aia if ia.access_method == AuthorityInformationAccessOID.OCSP
        ]
        if not ocsps:
            return True

        ocsp_url = ocsps[0].access_location.value
        builder = ocsp.OCSPRequestBuilder()

        # Hm, some OCSP servers do not support SHA256, so let's use SHA1 until we
        # are told otherwise.
        builder = builder.add_certificate(
            cert.parsed, issuer_cert.parsed, hashes.SHA1()
        )
        req = builder.build()
        start = time.time()
        try:
            if self._recorder.is_injecting():
                ocsp_resp = self._recorder.inject_response()

            else:
                ocsp_resp = requests.post(
                    ocsp_url,
                    headers={"Content-Type": "application/ocsp-request"},
                    data=req.public_bytes(serialization.Encoding.DER),
                    timeout=5,
                    proxies=self._proxies,
                )
                self._recorder.trace_response(
                    time.time() - start, recorder.SocketEvent.DATA, ocsp_resp
                )

        except requests.Timeout:
            self._recorder.trace_response(
                time.time() - start, recorder.SocketEvent.TIMEOUT
            )
            cert.ocsp_status = tls.OcspStatus.TIMEOUT
            cert.mark_untrusted(f"connection to OCSP server {ocsp_url} timed out")
            return False

        except Exception:
            self._recorder.trace_response(
                time.time() - start, recorder.SocketEvent.CLOSURE
            )
            cert.ocsp_status = tls.OcspStatus.INVALID_RESPONSE
            cert.mark_untrusted(f"connection to OCSP server {ocsp_url} failed")
            return False

        if ocsp_resp.ok:
            cert.ocsp_status = self._verify_ocsp_resp(
                cert, ocsp_resp.content, self._recorder.get_timestamp(), issuer_cert,
            )
            issue = f"OCSP status is {cert.ocsp_status}"
            if cert.ocsp_status is not tls.OcspStatus.NOT_REVOKED:
                cert.mark_untrusted(issue)
                return False

            else:
                logging.debug(f"{issue} for certificate {cert}")
                return True

        return False

    def _issuer_certs(self, cert):
        """Get a list of all potential issuers from the certificate chain."""
        issuers = []
        for issuer_idx in range(1, len(self.certificates)):
            issuer_cert = self.certificates[issuer_idx]
            if cert.auth_key_id and issuer_cert.subject_key_id:
                match = cert.auth_key_id == issuer_cert.subject_key_id

            else:
                match = cert_utils.equal_names(
                    cert.parsed.issuer, issuer_cert.parsed.subject
                )

            if match:
                issuers.append((issuer_idx, issuer_cert))

        return issuers

    def _determine_trust_path(self, cert, idx, timestamp, domain_name, full_validation):
        trust_path = [idx] if idx is not None else []
        if cert.trusted is not tls.ScanState.UNDETERMINED:
            return trust_path

        cert.trusted = tls.ScanState.TRUE
        if not cert.has_valid_period(timestamp) and not full_validation:
            return trust_path

        if idx == 0:
            if not cert.has_valid_subject(domain_name) and not full_validation:
                return trust_path

        elif idx is None:
            cert.issuer_cert = cert
            return trust_path

        if cert.self_signed:
            issuers = [(idx, cert)]
            if not self._trust_store.cert_in_trust_store(cert):
                cert.mark_untrusted("self-signed certificate not found in trust store")
                if not full_validation:
                    return trust_path

            else:
                self.root_cert_transmitted = True
                self.root_cert = None

        else:
            issuers = self._issuer_certs(cert)
            # placeholder for certificate from the trust store
            issuers.append((None, None))

        issuer_trust_path = []
        for issuer_idx, issuer_cert in issuers:
            if not issuer_cert:
                issuer_cert = self._trust_store.issuer_in_trust_store(
                    cert.parsed.issuer
                )
                if issuer_cert:
                    issuer_cert.from_trust_store = True
                    self.root_cert_transmitted = False
                    self.root_cert = issuer_cert

                else:
                    if len(issuers) == 1:
                        cert.mark_untrusted(
                            f"issuer certificate "
                            f'"{cert.parsed.issuer.rfc4514_string()}" not found '
                            f"in trust store"
                        )
                    break

            if not cert.self_signed:
                issuer_trust_path = self._determine_trust_path(
                    issuer_cert, issuer_idx, timestamp, domain_name, full_validation,
                )
                if issuer_cert.trusted is tls.ScanState.FALSE:
                    continue

            else:
                issuer_trust_path = []

            cert.issuer_cert = issuer_cert

            try:
                issuer_cert.validate_cert_signature(cert)

            except Exception:
                cert.mark_untrusted("invalid signature")
                continue

            if (
                not self._valid_crl(cert, issuer_cert, timestamp)
                and not full_validation
            ):
                continue

            if (
                not self._valid_ocsp(cert, issuer_cert, timestamp)
                and not full_validation
            ):
                continue

            if issuer_idx is not None:
                if issuer_idx < idx:
                    self.issues.append("certificates of the chain are not in sequence")

            break

        return trust_path + issuer_trust_path

    def validate(
        self, timestamp: datetime.datetime, domain_name: str, raise_on_failure: bool
    ) -> None:
        """Validates the certificate chain. Only the minimal checks are supported.

        If a discrepancy is found, an exception is raised (depending on
        raise_on_failure).

        Arguments:
            timestamp: the timestamp to check against
            domain_name: the domain name to validate the host certificate against
            raise_on_failure: whether an exception shall be raised if the
                validation fails or not. Useful for a TLS scan, as the scan
                shall continue.

        Raises:
            tls.UntrustedCertificate: in case a certificate within the chain cannot be
                validated and `raise_on_failure` is True.
        """

        server_cert = self.certificates[0]
        valid = self._cert_chain_cache.get_cached_validation_state(self)
        if valid is not None:
            logging.debug(
                f"using certificate chain validation status {valid} from cache"
            )
            if not valid and raise_on_failure:
                raise tls.UntrustedCertificate(
                    f"cached status for {self.certificates[0]} is not valid"
                )
            return

        trust_path = self._determine_trust_path(
            server_cert, 0, timestamp, domain_name, not raise_on_failure
        )

        self.successful_validation = all(
            [self.certificates[idx].trusted is tls.ScanState.TRUE for idx in trust_path]
        )
        self._cert_chain_cache.update_cached_validation_state(self)

        if raise_on_failure and not self.successful_validation:
            issue = "not trusted"
            for idx in trust_path:
                if self.certificates[idx].issues:
                    issue = self.certificates[idx].issues[0]
                    break

            raise tls.UntrustedCertificate(f"certificate {server_cert}: {issue}")

        # And now check for gratuitous certificate in the chain
        if not raise_on_failure and trust_path:
            for idx, cert in enumerate(self.certificates):
                if idx not in trust_path:
                    if cert.trusted is tls.ScanState.UNDETERMINED:
                        cert.issues.append(
                            "gratuitous certificate, not part of trust chain"
                        )

                    elif cert.trusted is tls.ScanState.FALSE:
                        cert.issues.append(
                            "certificate part of untrusted alternate trust path"
                        )

    def serialize(self) -> List[str]:
        """Serialize the certificate chain

        Returns:
            A list of certificates which build the chain. The format is a str,
            representing the DER-format for each certificate.
        """
        return [cert.bytes.hex() for cert in self.certificates]

    def deserialize(self, chain: List[str]) -> None:
        """Deserializes a certificate chain.

        Arguments:
            chain (list of str): The list of certificates of the chain. Each certificate
                is represented in DER-format as a string.
        """
        for cert in chain:
            self.append_bin_cert(bytes.fromhex(cert))


class CertChainCache(object):
    """Caches the validation state for a given certificate chain
    """

    _CACHE_SIZE = 100

    def __init__(self):
        self._cache = {}

    def get_cached_validation_state(self, cert_chain: CertChain) -> bool:
        """Returns the validation state of a certificate chain from the cache.

        If not present in the cache, the certificate chain is added with the status
        False.

        Arguments:
            cert_chain: the certificate chain object

        Returns:
            the validation status of the certificate chain from the cache or
            None if the certificate chain is not found.
        """

        val = self._cache.get(cert_chain.digest)
        if val is None:
            # Here keep cache at a reasonable size
            if len(self._cache) >= self._CACHE_SIZE:
                del self._cache[next(iter(self._cache))]

            self._cache[cert_chain.digest] = False

        return val

    def update_cached_validation_state(self, cert_chain: CertChain) -> None:
        """Updates a certificate chain entry in the cache.

        Arguments:
            cert_chain: the certificate chain object
        """

        self._cache[cert_chain.digest] = cert_chain.successful_validation



from tlsmate.plugin import Worker, Plugin, Args, BaseCommand
from tlsmate import tls
from tlsmate.plugins.scan import ArgPort, ArgSni, ArgHost
from tlsmate.structs import ConfigItem


class CipherPrinterWorker(Worker):
    name = "cipher printer"
    descr = "executing a single handshake and print the negotiated cipher suite"
    prio = 100

    def run(self):
        # Let's use a default client profile which has a high probability to
        # successfully interoperate with a typical web server.
        self.client.set_profile(tls.Profile.INTEROPERABILITY)

        # Now open a TLS connection and execute a typical TLS handshake. Print the
        # cipher suite selected by the server.
        with self.client.create_connection() as conn:
            conn.handshake()
            print(f"{self.config.get('text')}: {conn.msg.server_hello.cipher_suite}")


class ArgText(Plugin):
    """Argument for the text
    """

    config = ConfigItem("text", type=str, default="cipher suite")
    cli_args = Args(
        "--text",
        help="the text used to print the negotiated cipher suite",
        default=None,
    )


@BaseCommand.extend
class SubcommandCipherPrinter(Plugin):
    subcommand = Args("cipher-printer", help="prints the negotiated cipher suite")
    plugins = [ArgPort, ArgSni, ArgText, ArgHost]
    workers = [CipherPrinterWorker]



# -*- coding: utf-8 -*-
"""Module containing the class implementing hashing functions & prf
"""
# import basic stuff
import abc
import math
import struct
import logging
from typing import Optional

# import own stuff
import tlsmate.pdu as pdu

# import other stuff
from cryptography.hazmat.primitives import hashes, hmac
from cryptography.hazmat.primitives.kdf.hkdf import HKDFExpand


class _Backend(metaclass=abc.ABCMeta):
    """Base class for backend implementations

    Arguments:
        algo (:obj:`cryptography.hazmat.primitives.hashes.HashPrimitive`): The hash
            algorithm to be used in the backend.
    """

    @abc.abstractmethod
    def __init__(self, algo: Optional[hashes.Hash]):
        raise NotImplementedError

    @staticmethod
    def _hmac_func(secret, msg, hash_algo):
        """Provides HMAC function.

        Arguments:
            secret (bytes): The secret used for the HMAC.
            msg (bytes): The array of bytes for which the HMAC shall be generated.
            hash_algo (:obj:`cryptography.hazmat.primitives.hashes.HashPrimitive`): The
                hash algorithm used in the HMAC function.

        Returns:
            bytes: The calculated HMAC value, having the same length than the output
            of the hash algorithm.
        """

        hmac_object = hmac.HMAC(secret, hash_algo())
        hmac_object.update(msg)
        return hmac_object.finalize()

    @staticmethod
    def _expand(secret, seed, size, hash_algo):
        """Provides an expand funtion.

        This function is used to expand a key with high entropy to a byte string of
        the desired length.

        Arguments:
            secret (bytes): The key used for the expand function.
            seed (bytes): The seed for the function.
            size (int): The desired number of bytes for the output
            hash_algo (:obj:`cryptography.hazmat.primitives.hashes.HashPrimitive`): The
                hash algorithm used in the underlying HMAC function.

        Returns:
            (bytes): An bytearray of the desired length.
        """

        out = b""
        ax = bytes(seed)
        while len(out) < size:
            ax = __class__._hmac_func(secret, ax, hash_algo)
            out = out + __class__._hmac_func(secret, ax + seed, hash_algo)

        return out[:size]

    @abc.abstractmethod
    def update_msg_digest(self, msg):
        """Function to add a message to a message digest.

        Arguments:
            msg (bytes): The message to add.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def current_msg_digest(self):
        """Determine the current value of the message digest.

        Returns:
            bytes: The calculated current  message digest, output has the same length
            than the hash length.
        """

        raise NotImplementedError

    @abc.abstractmethod
    def prf(self, secret, label, seed, size):
        """Provide a PRF.

        Arguments:
            secret (bytes): The key used for the expand function.
            label (bytes): The label as a bytestring.
            seed (bytes): The seed for the function.
            size (int): The desired number of bytes for the output

        Returns:
            (bytes): An bytearray of the desired length.
        """

        raise NotImplementedError


class _BackendTls10(_Backend):
    """Implements a backend for TLS1.0 and TLS1.1.

    The message digest is using a fixed combination of MD5 and SHA.
    """

    def __init__(self, hash_algo):
        self._msg_digest_md5 = hashes.Hash(hashes.MD5())
        self._msg_digest_sha = hashes.Hash(hashes.SHA1())

    def update_msg_digest(self, msg):
        self._msg_digest_md5.update(msg)
        self._msg_digest_sha.update(msg)

    def current_msg_digest(self):
        tmp_digest_md5 = self._msg_digest_md5.copy()
        tmp_digest_sha = self._msg_digest_sha.copy()
        return tmp_digest_md5.finalize() + tmp_digest_sha.finalize()

    def prf(self, secret, label, seed, size):
        length = math.ceil(len(secret) / 2)
        s1_md5 = secret[:length]
        s2_sha = secret[-length:]
        md5_bytes = self._expand(s1_md5, label + seed, size, hashes.MD5)
        sha_bytes = self._expand(s2_sha, label + seed, size, hashes.SHA1)
        result = bytearray(sha_bytes)
        for i, b in enumerate(md5_bytes):
            result[i] ^= b

        return result

    def ssl30_digest(self, master_secret, label):
        md5 = self._msg_digest_md5.copy()
        sha = self._msg_digest_sha.copy()
        md5.update(label + master_secret + (b"\x36" * 48))
        sha.update(label + master_secret + (b"\x36" * 40))
        md5_2 = hashes.Hash(hashes.MD5())
        sha_2 = hashes.Hash(hashes.SHA1())
        md5_2.update(master_secret + (b"\x5c" * 48) + md5.finalize())
        sha_2.update(master_secret + (b"\x5c" * 40) + sha.finalize())
        return md5_2.finalize() + sha_2.finalize()

    def ssl3_master_secret(self, pms, randoms):
        sha = hashes.Hash(hashes.SHA1())
        sha.update(b"A" + pms + randoms)
        md5 = hashes.Hash(hashes.MD5())
        md5.update(pms + sha.finalize())
        ms = md5.finalize()

        sha = hashes.Hash(hashes.SHA1())
        sha.update(b"BB" + pms + randoms)
        md5 = hashes.Hash(hashes.MD5())
        md5.update(pms + sha.finalize())
        ms += md5.finalize()

        sha = hashes.Hash(hashes.SHA1())
        sha.update(b"CCC" + pms + randoms)
        md5 = hashes.Hash(hashes.MD5())
        md5.update(pms + sha.finalize())
        return ms + md5.finalize()

    def ssl3_key_material(self, ms, randoms, size):
        block = b""
        for idx in range((size + 15) // 16):
            prefix = bytes([65 + idx] * (idx + 1))
            sha = hashes.Hash(hashes.SHA1())
            sha.update(prefix + ms + randoms)
            md5 = hashes.Hash(hashes.MD5())
            md5.update(ms + sha.finalize())
            block += md5.finalize()

        return block


class _BackendTls12(_Backend):
    """Implements a backend for TLS1.2 and TLS1.3
    """

    def __init__(self, hash_algo):
        self._hash_algo = hash_algo
        self._msg_digest = hashes.Hash(hash_algo())

    def update_msg_digest(self, msg):
        self._msg_digest.update(msg)

    def current_msg_digest(self):
        return self._msg_digest.copy().finalize()

    def prf(self, secret, label, seed, size):
        return self._expand(secret, label + seed, size, self._hash_algo)

    def hkdf_extract(self, secret, salt):
        """TLS1.3 specific, implements the "hkdf_extract" function.

        Arguments:
            secret (bytes): The secret used for the extract function.
            salt (bytes): The salt for the extract function.

        Returns:
            bytes: The byte sequence generated having the same length than the
            underlying hash algorithm.
        """

        if secret is None:
            secret = b"\0" * self._hash_algo.digest_size

        h = hmac.HMAC(salt, self._hash_algo())
        h.update(secret)
        return h.finalize()

    def _hkdf_expand(self, secret, label, length):
        """TLS1.3 specific, implements the "hkdf_expand" function.

        This function is used to derive keying material as long as needed from a
        strong secret.

        Arguments:
            secret (bytes): The secret used for the expand function.
            label (bytes): The label as a bytestring.
            length (int): The desired number of bytes for the output

        Returns:
            bytes: A bytearray as long as specified by the length parameter.
        """

        hkdf = HKDFExpand(algorithm=self._hash_algo(), length=length, info=label)
        return hkdf.derive(secret)

    def hkdf_expand_label(self, secret, label, context, length):
        """TLS1.3 specific, implements the "hkdf_expand" function.

        This function is used to derive keying material as long as needed from a
        strong secret.

        Arguments:
            secret (bytes): The secret used for the expand function.
            label (bytes): The label as a bytestring.
            context (bytes): handshake context, e.g. the message digest of dedicated
                messages.
            length (int): The desired number of bytes for the output

        Returns:
            bytes: A bytearray as long as specified by the length parameter.
        """

        label_bytes = ("tls13 " + label).encode()
        hkdf_label = (
            struct.pack("!H", length)
            + struct.pack("!B", len(label_bytes))
            + label_bytes
            + struct.pack("!B", len(context))
            + context
        )
        return self._hkdf_expand(secret, hkdf_label, length)


class Kdf(object):
    """Implements several cryptographic functions, mainly intended to derive keys.
    """

    def __init__(self) -> None:
        self._backend: Optional[_Backend] = None

    def start_msg_digest(self) -> None:
        """Get ready to add messages to the message digest

        We start with the client hello, but backend is not defined yet.
        """
        self._msg_digest_queue: Optional[bytearray] = None
        self._msg_digest = None
        self._msg_digest_active = True
        self._backend = None
        self._empty_msg_digest: Optional[bytes] = None
        self._all_msgs = bytearray()

    def set_msg_digest_algo(self, hash_algo: Optional[hashes.Hash]) -> None:
        """Function to set the hash algo for the message digest.

        Note:
            The message digest starts with the ClientHello, but the used hash
            algorithm is only determined when the ServerHello is received, which
            means at least the ClientHello must be queued.

        Arguments:
            hash_algo: The negotiated hash algorithm implementation, determined
                with the reception of the ServerHello.
        """

        if hash_algo is None:
            self._backend = _BackendTls10(hash_algo)

        else:
            self._backend = _BackendTls12(hash_algo)

        self._empty_msg_digest = self.current_msg_digest()
        if self._msg_digest_queue is not None:
            self._backend.update_msg_digest(self._msg_digest_queue)
            self._msg_digest_queue = None

    def update_msg_digest(self, msg: bytes) -> None:
        """Add a message to the message digest.

        Note:
            The message might be queued internally until the hash algorithm is
            determined (reception of the ServerHello).

        Arguments:
            msg: The message to add.
        """

        if not self._msg_digest_active:
            return

        logging.debug(f"add to msg_digest: {pdu.dump_short(msg, start=20)}")
        self._all_msgs.extend(msg)
        if self._backend is None:
            if self._msg_digest_queue is None:
                self._msg_digest_queue = bytearray(msg)

            else:
                self._msg_digest_queue.extend(msg)

        else:
            self._backend.update_msg_digest(msg)

    def get_handshake_messages(self) -> bytearray:
        """Get all messages received/sent so far.

        Returns:
            the concatenation of all messages.
        """

        return self._all_msgs

    def empty_msg_digest(self) -> Optional[bytes]:
        """Return the message digest for an emty message array.

        Returns:
            The message digest for no given messages.
        """

        return self._empty_msg_digest

    def current_msg_digest(self, suspend: bool = False) -> Optional[bytes]:
        """Gets the message digest.

        Arguments:
            suspend: If set to False, provide the message digest in the current
                state, but allow to add more messages to the digest later on.
                If set to True, provide the message digest, but do not consider
                further messages.

        Returns:
            The calculated message digest, output has the same length than the
            hash length.
        """

        if suspend:
            self._msg_digest_active = False

        return self._backend.current_msg_digest()  # type: ignore

    def msg_digest_active(self) -> bool:
        """Returns the state of the message digest.

        Returns:
            an indication if the message digest is suspended or active.
        """

        return self._msg_digest_active

    def resume_msg_digest(self) -> None:
        """Change the state of the message digest to active
        """

        self._msg_digest_active = True

    def prf(self, secret: bytes, label: bytes, seed: bytes, size: int) -> bytes:
        """Implements a pseudo random function.

        Arguments:
            secret: The secret used for the PRF.
            label: The label as a bytestring.
            seed: The seed for the function.
            size: The desired number of bytes for the output

        Returns:
            A bytearray as long as specified by the size parameter.
        """

        return self._backend.prf(secret, label, seed, size)  # type: ignore

    def hkdf_extract(self, secret: bytes, salt: bytes) -> bytes:
        """HKDF-extract function for TLS1.3

        Arguments:
            secret: The secret used for the extract function.
            salt: The salt for the extract function.

        Returns:
            The byte sequence generated having the same length than the hash
            algorithm used in the underlying hash function.
        """

        return self._backend.hkdf_extract(secret, salt)  # type: ignore

    def hkdf_expand_label(
        self, secret: bytes, label: bytes, msg_digest: bytes, length: int
    ) -> bytes:
        """HKDF-expand-label function for TLS1.3

        Arguments:
            secret: The secret used for the expand function.
            label: The label as a bytestring.
            msg_digest: The message digest value having the same length than
                the underlying hash function.
            length: The desired number of bytes for the output

        Returns:
            A bytearray as long as specified by the length parameter.
        """

        return self._backend.hkdf_expand_label(  # type: ignore
            secret, label, msg_digest, length
        )



# -*- coding: utf-8 -*-
"""Implement unit tests for the module utils.
"""
from tlsmate import tls
from tlsmate import utils
from tlsmate.workers.base_vulnerabilities import ScanBaseVulnerabilities


def test_base_vulnerabilities_ok(tlsmate, server_profile):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.beast is tls.ScanState.FALSE
    assert prof_vuln.crime is tls.ScanState.FALSE
    assert prof_vuln.sweet_32 is tls.ScanState.FALSE
    assert prof_vuln.freak is tls.ScanState.FALSE
    assert prof_vuln.logjam is tls.Logjam.OK


def test_base_vulnerabilities_nok(tlsmate, server_profile_base_vuln):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_base_vuln)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.beast is tls.ScanState.TRUE
    assert prof_vuln.crime is tls.ScanState.TRUE
    assert prof_vuln.sweet_32 is tls.ScanState.TRUE
    assert prof_vuln.freak is tls.ScanState.TRUE
    assert prof_vuln.logjam is tls.Logjam.PRIME512


def test_base_vulnerabilities_no_compr(tlsmate, server_profile_no_compr):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_no_compr)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.crime is tls.ScanState.UNDETERMINED


def test_base_vulnerabilities_no_features(tlsmate, server_profile_no_features):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_no_features)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.crime is tls.ScanState.UNDETERMINED


def test_base_vulnerabilities_logjam_1024_common(tlsmate, server_profile_logjam_common):
    tlsmate.server_profile.load(
        utils.deserialize_data(str(server_profile_logjam_common))
    )
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.logjam is tls.Logjam.PRIME1024_COMMON


def test_base_vulnerabilities_logjam_1024_cust(tlsmate, server_profile_logjam_cust):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_logjam_cust)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.logjam is tls.Logjam.PRIME1024_CUSTOMIZED


def test_base_vulnerabilities_logjam_no_dh_group(tlsmate, server_profile_no_dh_group):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_no_dh_group)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.logjam is tls.Logjam.UNDETERMINED


def test_base_vulnerabilities_logjam_no_dh(tlsmate, server_profile_no_dh):
    tlsmate.server_profile.load(utils.deserialize_data(str(server_profile_no_dh)))
    ScanBaseVulnerabilities(tlsmate).run()
    prof_vuln = tlsmate.server_profile.vulnerabilities
    assert prof_vuln.logjam is tls.Logjam.NA



# -*- coding: utf-8 -*-
"""Module containing the test suite for ephemeral key reuse
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.server_profile as server_profile
import tlsmate.tls as tls
import tlsmate.utils as utils

# import other stuff

HANDSHAKE_COUNT = 5


def _determine_status(keys):
    if len(keys) == HANDSHAKE_COUNT:
        if len(set(keys)) == HANDSHAKE_COUNT:
            return tls.ScanState.FALSE

        else:
            return tls.ScanState.TRUE

    return tls.ScanState.UNDETERMINED


class ScanEphemeralKeyReuse(plg.Worker):
    name = "ephemeral_key_reuse"
    descr = "scan for ephemeral key reuse"
    prio = 31

    def _tls12_handshakes(self, cs, pub_key):
        if not cs:
            return tls.ScanState.NA

        self.client.profile.cipher_suites = cs
        keys = []
        for _ in range(HANDSHAKE_COUNT):
            with self.client.create_connection() as conn:
                conn.handshake()

            if conn.msg.server_key_exchange is not None:
                key = pub_key(conn.msg.server_key_exchange)
                if key in keys:
                    return tls.ScanState.TRUE

                keys.append(key)

        return _determine_status(keys)

    def _scan_tls12(self):
        versions = [
            tls.Version.SSL30,
            tls.Version.TLS10,
            tls.Version.TLS11,
            tls.Version.TLS12,
        ]
        prof_values = self.server_profile.get_profile_values(versions)
        if not prof_values.versions:
            return (tls.ScanState.NA, tls.ScanState.NA)

        self.client.init_profile(profile_values=prof_values)
        dhe_cs = utils.filter_cipher_suites(
            prof_values.cipher_suites,
            key_algo=[
                tls.KeyExchangeAlgorithm.DHE_DSS,
                tls.KeyExchangeAlgorithm.DHE_RSA,
            ],
        )
        ecdhe_cs = utils.filter_cipher_suites(
            prof_values.cipher_suites,
            key_algo=[
                tls.KeyExchangeAlgorithm.ECDHE_RSA,
                tls.KeyExchangeAlgorithm.ECDHE_ECDSA,
            ],
        )

        tls12_dhe = self._tls12_handshakes(
            dhe_cs, pub_key=lambda msg: msg.dh.public_key
        )
        tls12_ecdhe = self._tls12_handshakes(
            ecdhe_cs, pub_key=lambda msg: msg.ec.public
        )

        return tls12_dhe, tls12_ecdhe

    def _tls13_handshakes(self, shares):
        if not shares:
            return tls.ScanState.NA

        self.client.profile.supported_groups = shares
        self.client.profile.key_shares = shares
        keys = []
        for _ in range(HANDSHAKE_COUNT):
            with self.client.create_connection() as conn:
                conn.send(msg.ClientHello)
                server_hello = conn.wait(msg.ServerHello)
                key_share_ext = server_hello.get_extension(tls.Extension.KEY_SHARE)
                if not key_share_ext:
                    return tls.ScanState.UNDETERMINED

                key_share = key_share_ext.key_shares[0].key_exchange
                if key_share in keys:
                    return tls.ScanState.TRUE

                keys.append(key_share)
                self.client.profile.cipher_suites = [server_hello.cipher_suite]

        return _determine_status(keys)

    def _scan_tls13(self):
        prof_values = self.server_profile.get_profile_values([tls.Version.TLS13])
        if not prof_values.versions:
            return (tls.ScanState.NA, tls.ScanState.NA)

        self.client.init_profile(profile_values=prof_values)
        ecdhe_shares = []
        dhe_shares = []
        for share in prof_values.key_shares:
            if share.value > 255:
                dhe_shares.append(share)
            else:
                ecdhe_shares.append(share)

        return (
            self._tls13_handshakes(dhe_shares),
            self._tls13_handshakes(ecdhe_shares),
        )

    def run(self):

        tls12_dhe, tls12_ecdhe = self._scan_tls12()
        tls13_dhe, tls13_ecdhe = self._scan_tls13()

        self.server_profile.allocate_features()
        if not hasattr(self.server_profile.features, "ephemeral_key_reuse"):
            self.server_profile.features.ephemeral_key_reuse = (
                server_profile.SPEphemeralKeyReuse()
            )

        ekr = self.server_profile.features.ephemeral_key_reuse
        ekr.tls12_dhe_reuse = tls12_dhe
        ekr.tls12_ecdhe_reuse = tls12_ecdhe
        ekr.tls13_dhe_reuse = tls13_dhe
        ekr.tls13_ecdhe_reuse = tls13_ecdhe



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import pathlib
from tlsmate.workers.padding_oracle import ScanPaddingOracle
from tlsmate.tlssuite import TlsSuiteTester
from tlsmate.tlssuite import TlsLibrary


class TestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    sp_in_yaml = "profile_sig_algos_openssl1_0_2"
    recorder_yaml = "recorder_cbc_padding_oracle_high"
    path = pathlib.Path(__file__)
    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa --no-cert-chain "
        "--ca-file ca-certificates -- -www -cipher ALL"
    )
    library = TlsLibrary.openssl1_0_1e

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        server_profile = tlsmate.server_profile
        self.config.set("oracle_accuracy", "high")
        ScanPaddingOracle(tlsmate).run()
        profile = server_profile.make_serializable()
        assert profile["vulnerabilities"]["lucky_minus_20"] == "TRUE"
        assert profile["vulnerabilities"]["poodle"] == "TRUE"
        assert profile["vulnerabilities"]["tls_poodle"] == "FALSE"
        cbc = profile["vulnerabilities"]["cbc_padding_oracle"]
        assert cbc["accuracy"] == "HIGH"
        assert cbc["vulnerable"] == "TRUE"
        assert len(cbc["oracles"]) == 1
        oracle = cbc["oracles"][0]
        assert oracle["observable"] == "FALSE"
        assert oracle["strong"] == "FALSE"
        assert oracle["types"] == ["LUCKY_MINUS_20"]
        groups = oracle["cipher_group"]
        assert any([x["record_protocol"]["name"] == "APPLICATION_DATA" for x in groups])
        assert any([x["record_protocol"]["name"] == "HANDSHAKE" for x in groups])
        assert any([x["record_protocol"]["name"] == "ALERT" for x in groups])
        assert any([x["version"]["name"] == "TLS10" for x in groups])
        assert any([x["version"]["name"] == "TLS11" for x in groups])
        assert any([x["version"]["name"] == "TLS12" for x in groups])


if __name__ == "__main__":
    TestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module scanning for heartbeat support
"""
# import basic stuff

# import own stuff
import tlsmate.msg as msg
import tlsmate.plugin as plg
import tlsmate.tls as tls

# import other stuff


class ScanHeartbeat(plg.Worker):
    name = "heartbeat"
    descr = "scan for heartbeat support"
    prio = 30

    def run(self):
        state = tls.HeartbeatState.UNDETERMINED
        versions = [
            tls.Version.TLS10,
            tls.Version.TLS11,
            tls.Version.TLS12,
            tls.Version.TLS13,
        ]
        prof_values = self.server_profile.get_profile_values(versions, full_hs=True)
        if not prof_values.versions:
            state = tls.HeartbeatState.NA

        else:
            self.client.init_profile(profile_values=prof_values)
            self.client.profile.heartbeat_mode = tls.HeartbeatMode.PEER_ALLOWED_TO_SEND
            with self.client.create_connection() as conn:
                conn.handshake()
                if conn.handshake_completed:
                    hb = conn.msg.server_hello.get_extension(tls.Extension.HEARTBEAT)
                    if hb:
                        if hb.heartbeat_mode is tls.HeartbeatMode.PEER_ALLOWED_TO_SEND:
                            req = msg.HeartbeatRequest()
                            req.payload = b"abracadabra"
                            req.payload_length = len(req.payload)
                            req.padding = b"\xff" * 16
                            conn.send(req)
                            rec_msg = conn.wait(
                                msg.Any, timeout=2000, fail_on_timeout=False,
                            )
                            if rec_msg is None:
                                state = tls.HeartbeatState.NOT_REPONDING

                            elif isinstance(rec_msg, msg.HeartbeatResponse):
                                if (
                                    req.payload_length == rec_msg.payload_length
                                    and req.payload == rec_msg.payload
                                ):
                                    state = tls.HeartbeatState.TRUE

                                else:
                                    state = tls.HeartbeatState.WRONG_RESPONSE

                            else:
                                state = tls.HeartbeatState.UNEXPECTED_MESSAGE

                        else:
                            state = tls.HeartbeatState.FALSE

                    else:
                        state = tls.HeartbeatState.FALSE

        self.server_profile.allocate_features()
        self.server_profile.features.heartbeat = state



# -*- coding: utf-8 -*-
"""Implement unit tests for the module utils.
"""
from tlsmate import tls
from tlsmate import msg
from tlsmate import ext
from tlsmate.socket import Socket
from tlsmate.connection import TlsConnection
from tlsmate.workers.heartbeat import ScanHeartbeat
from tlsmate.server_profile import ServerProfile
from tlsmate.client import Client


class Object(object):
    pass


def get_profile_values(self, filter_versions, full_hs=False):
    prof = Object()
    prof.versions = [True]
    return prof


def init_profile(self, **kwargs):
    pass


def open_socket(self, addr):
    pass


def heartbeat_response_ok(self, timeout=5):
    return bytes.fromhex(
        "18 03 03 00 2e 02 00 0b 61 62 72 61 63 61 64 61 62 72 61 "
        "00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f"
    )


def handshake_ok(self):
    self.handshake_completed = True
    server_hello = msg.ServerHello()
    hb = ext.ExtHeartbeat(heartbeat_mode=tls.HeartbeatMode.PEER_ALLOWED_TO_SEND)
    server_hello.extensions.append(hb)
    self.msg.server_hello = server_hello


def test_heartbeat_ok(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(Socket, "recv_data", heartbeat_response_ok)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_ok)
    ScanHeartbeat(tlsmate).run()
    assert tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.TRUE


def heartbeat_response_wrong_resp(self, timeout=5):
    return bytes.fromhex(
        "18 03 03 00 2e 02 00 0b 62 62 72 61 63 61 64 61 62 72 61 "
        "00 01 02 03 04 05 06 07 08 09 0a 0b 0c 0d 0e 0f"
    )


def test_heartbeat_wrong_resp(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(Socket, "recv_data", heartbeat_response_wrong_resp)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_ok)
    ScanHeartbeat(tlsmate).run()
    assert (
        tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.WRONG_RESPONSE
    )


def heartbeat_response_wrong_msg(self, timeout=5):
    return bytes.fromhex("15 03 03 00 02 01 00")


def test_heartbeat_wrong_msg(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(Socket, "recv_data", heartbeat_response_wrong_msg)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_ok)
    ScanHeartbeat(tlsmate).run()
    assert (
        tlsmate.server_profile.features.heartbeat
        is tls.HeartbeatState.UNEXPECTED_MESSAGE
    )


def handshake_hb_not_allowed(self):
    self.handshake_completed = True
    server_hello = msg.ServerHello()
    hb = ext.ExtHeartbeat(heartbeat_mode=tls.HeartbeatMode.PEER_NOT_ALLOWED_TO_SEND)
    server_hello.extensions.append(hb)
    self.msg.server_hello = server_hello


def test_heartbeat_not_allowed(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_hb_not_allowed)
    ScanHeartbeat(tlsmate).run()
    assert tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.FALSE


def handshake_no_ext(self):
    self.handshake_completed = True
    self.msg.server_hello = msg.ServerHello()


def test_heartbeat_no_ext(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_no_ext)
    ScanHeartbeat(tlsmate).run()
    assert tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.FALSE


def get_profile_values_no_versions(self, filter_versions, full_hs=False):
    prof = Object()
    prof.versions = []
    return prof


def test_no_versions(monkeypatch, tlsmate):
    monkeypatch.setattr(
        ServerProfile, "get_profile_values", get_profile_values_no_versions
    )
    ScanHeartbeat(tlsmate).run()
    assert tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.NA


def heartbeat_no_response(self, timeout=5):
    raise tls.TlsMsgTimeoutError


def test_heartbeat_no_resonse(monkeypatch, tlsmate):
    monkeypatch.setattr(ServerProfile, "get_profile_values", get_profile_values)
    monkeypatch.setattr(Client, "init_profile", init_profile)
    monkeypatch.setattr(Socket, "open_socket", open_socket)
    monkeypatch.setattr(Socket, "recv_data", heartbeat_no_response)
    monkeypatch.setattr(TlsConnection, "handshake", handshake_ok)
    ScanHeartbeat(tlsmate).run()
    assert tlsmate.server_profile.features.heartbeat is tls.HeartbeatState.NOT_REPONDING



# -*- coding: utf-8 -*-
"""Implements a class to test the text server profile worker.
"""
import pathlib
import sys

from tlsmate.workers.server_profile import ReadProfileWorker
from tlsmate.workers.text_server_profile import TextProfileWorker
from tlsmate.config import Configuration
from tlsmate.tlsmate import TlsMate
from tlsmate.structs import ConfigItem
from tlsmate import command
from tlsmate.tlssuite import TlsSuiteTester, TlsLibrary

RECORDINGS_PATH = pathlib.Path(__file__).resolve().parent / "recordings"
SERVER_PROFILE = RECORDINGS_PATH / "profile_text_ssl2.yaml"


class NoTestCase(TlsSuiteTester):
    """Class used for tests with pytest.

    For more information refer to the documentation of the TcRecorder class.
    """

    server_cmd = (
        "utils/start_openssl --version {library} --port {server_port} "
        "--cert1 server-rsa --cert2 server-ecdsa --no-cert-chain "
        "--ca-file ca-certificates -- -www -cipher ALL -ssl2"
    )
    library = TlsLibrary.openssl1_0_1e

    server = "localhost"

    def run(self, tlsmate, is_replaying):
        cmd = (
            f"tlsmate scan localhost --port={self.config.get('port')} "
            f"--format=yaml --write-profile={SERVER_PROFILE}"
        )
        sys.argv = cmd.split()
        command.main()


def test_server_profile(style_file):
    config = Configuration()
    config.register(ConfigItem("read_profile"))
    config.register(ConfigItem("style", type=str, default=str(style_file)))
    config.register(ConfigItem("color", type=bool, default=True))
    config.set("read_profile", str(SERVER_PROFILE))
    tlsmate = TlsMate(config)
    ReadProfileWorker(tlsmate).run()
    TextProfileWorker(tlsmate).run()


if __name__ == "__main__":
    NoTestCase().entry(is_replaying=False)



# -*- coding: utf-8 -*-
"""Module containing the TLS Extension classes
"""

# import basic stuff
import abc
import time
import logging
from typing import Tuple, Any, Optional, List, Union

# import own stuff
import tlsmate.client_state as client_state
import tlsmate.tls as tls
import tlsmate.structs as structs
import tlsmate.pdu as pdu

# import other stuff


class Extension(metaclass=abc.ABCMeta):
    """Abstract class for all extensions.
    """

    extension_id: tls.Extension
    """The extension id
    """

    @abc.abstractmethod
    def _serialize_ext_body(self, session: client_state.SessionState) -> bytes:
        """Serializes the content of an extension, i.e. excluding the header.

        Arguments:
            session: the session object that will be used to send this
                extensions.

        Returns:
            bytes: The serialized extensions body.
        """

        pass

    def serialize(self, session: client_state.SessionState) -> bytes:
        """Serializes an extensions, including the extensions header.

        Arguments:
            session: the session object that will be used to send this
                extensions.

        Returns:
            The serialized extension.
        """

        ext_body = self._serialize_ext_body(session)
        if self.extension_id is tls.Extension.UNKNOW_EXTENSION:
            ext_id = self.id  # type: ignore

        else:
            ext_id = self.extension_id.value

        ext = bytearray(pdu.pack_uint16(ext_id))
        ext.extend(pdu.pack_uint16(len(ext_body)))
        ext.extend(ext_body)
        return ext

    @staticmethod
    def deserialize(fragment: bytes, offset: int) -> Tuple["Extension", int]:
        """Deserializes an extension.

        Arguments:
            fragment: A PDU buffer as received from the network.
            offset: The offset within the fragment where the extension starts.

        Returns:
            The deserialized extension as an python object and the new offset
            into the fragment.
        """
        ext_id_int, offset = pdu.unpack_uint16(fragment, offset)
        ext_id = tls.Extension.val2enum(ext_id_int)
        ext_len, offset = pdu.unpack_uint16(fragment, offset)
        ext_body, offset = pdu.unpack_bytes(fragment, offset, ext_len)
        cls_name = deserialization_map.get(ext_id)
        if cls_name is None:
            return ExtUnknownExtension(id=ext_id_int, bytes=fragment[4:]), len(fragment)

        extension = cls_name()
        extension._deserialize_ext_body(ext_body)
        return extension, offset

    @abc.abstractmethod
    def _deserialize_ext_body(self, fragment):
        """Deserializes the body of an extension.

        Arguments:
            fragment (bytes): the body of the extension in binary format
        """


class ExtUnknownExtension(Extension):
    """Any extensions which is not known by tlsmate (yet).

    Arguments:
        id: the Id of the extensions as used in the PDU
        bytes: the content of the extension as a byte string
    """

    extension_id = tls.Extension.UNKNOW_EXTENSION

    def __init__(self, **kwargs: Any) -> None:
        self.id: Optional[int] = kwargs.get("id")
        self.bytes: Optional[bytes] = kwargs.get("bytes")

    def _serialize_ext_body(self, session):
        return self.bytes

    def _deserialize_ext_body(self, fragment):
        # deserialization implemented in base class.
        pass


class ExtServerNameIndication(Extension):
    """Represents the ServerNameIndication extension.

    Attributes:
        host_name: The name which identifies the host.
    """

    extension_id = tls.Extension.SERVER_NAME
    """:obj:`tlsmate.tls.Extension.SERVER_NAME`
    """

    def __init__(self, host_name: Optional[str] = None) -> None:
        self.host_name = host_name

    def _serialize_ext_body(self, session):
        # we only support exacly one list element: host_name
        ext = bytearray(pdu.pack_uint8(0))  # host_name
        ext.extend(pdu.pack_uint16(len(self.host_name)))
        ext.extend(pdu.pack_str(self.host_name))
        ext_body = bytearray(pdu.pack_uint16(len(ext)))
        ext_body.extend(ext)
        return ext_body

    def _deserialize_ext_body(self, fragment):
        if not len(fragment):
            return

        list_length, offset = pdu.unpack_uint16(fragment, 0)
        if offset + list_length != len(fragment):
            raise tls.ServerMalfunction(
                tls.ServerIssue.EXTENTION_LENGHT_ERROR, extension=self.extension_id
            )

        while offset < len(fragment):
            name_type, offset = pdu.unpack_uint8(fragment, offset)
            name_length, offset = pdu.unpack_uint16(fragment, offset)
            name, offset = pdu.unpack_bytes(fragment, offset, name_length)
            if name_type == 0:
                self.host_name = name.decode()

        if self.host_name is None:
            raise tls.ServerMalfunction(
                tls.ServerIssue.SNI_NO_HOSTNAME, extension=self.extension_id
            )


class ExtExtendedMasterSecret(Extension):
    """Represents the ExtendedMasterSecret extension.
    """

    extension_id = tls.Extension.EXTENDED_MASTER_SECRET
    """:obj:`tlsmate.tls.Extension.EXTENDED_MASTER_SECRET`
    """

    def _serialize_ext_body(self, session):
        return b""

    def _deserialize_ext_body(self, ext_body):
        if ext_body:
            raise tls.ServerMalfunction(
                tls.ServerIssue.EXTENTION_LENGHT_ERROR, extension=self.extension_id
            )


class ExtEncryptThenMac(Extension):
    """Represents the EncryptThenMac extension.
    """

    extension_id = tls.Extension.ENCRYPT_THEN_MAC
    """:obj:`tlsmate.tls.Extension.ENCRYPT_THEN_MAC`
    """

    def _serialize_ext_body(self, session):
        return b""

    def _deserialize_ext_body(self, ext_body):
        if ext_body:
            raise tls.ServerMalfunction(
                tls.ServerIssue.EXTENTION_LENGHT_ERROR, extension=self.extension_id
            )


class ExtRenegotiationInfo(Extension):
    """Represents the RenegotiationInfo extension.

    Attributes:
        renegotiated_connection: The opaque bytes.
    """

    extension_id = tls.Extension.RENEGOTIATION_INFO
    """:obj:`tlsmate.tls.Extension.RENEGOTIATION_INFO`
    """

    def __init__(self, renegotiated_connection: bytes = b"\0") -> None:
        self.renegotiated_connection = renegotiated_connection

    def _serialize_ext_body(self, session):
        ext_body = (
            pdu.pack_uint8(len(self.renegotiated_connection))
            + self.renegotiated_connection
        )
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        length, offset = pdu.unpack_uint8(ext_body, 0)
        self.renegotiated_connection, offset = pdu.unpack_bytes(
            ext_body, offset, length
        )


class ExtEcPointFormats(Extension):
    """Represents the EcPointFormat extension.

    Attributes:
        ec_point_formats: The list of supported point formats.
    """

    extension_id = tls.Extension.EC_POINT_FORMATS
    """:obj:`tlsmate.tls.Extension.EC_POINT_FORMATS`
    """

    def __init__(
        self, ec_point_formats: Optional[List[Union[tls.EcPointFormat, int]]] = None
    ) -> None:
        self.ec_point_formats = (
            ec_point_formats if ec_point_formats else [tls.EcPointFormat.UNCOMPRESSED]
        )

    def _serialize_ext_body(self, session):
        format_list = bytearray()
        for point_format in self.ec_point_formats:
            if type(point_format) == int:
                format_list.extend(pdu.pack_uint8(point_format))

            else:
                format_list.extend(pdu.pack_uint8(point_format.value))

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint8(len(format_list)))
        ext_body.extend(format_list)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        self.ec_point_formats = []
        length, offset = pdu.unpack_uint8(ext_body, 0)
        if offset + length != len(ext_body):
            raise tls.ServerMalfunction(
                tls.ServerIssue.EXTENTION_LENGHT_ERROR, extension=self.extension_id
            )

        for i in range(length):
            point_format, offset = pdu.unpack_uint8(ext_body, offset)
            self.ec_point_formats.append(tls.EcPointFormat.val2enum(point_format))


class ExtSupportedGroups(Extension):
    """Represents the SupportedGroup extension.

    Attributes:
        supported_groups: The list of supported groups.
    """

    extension_id = tls.Extension.SUPPORTED_GROUPS
    """:obj:`tlsmate.tls.Extension.SUPPORTED_GROUPS`
    """

    def __init__(
        self, supported_groups: Optional[List[tls.SupportedGroups]] = None
    ) -> None:
        self.supported_groups = supported_groups if supported_groups else []

    def _serialize_ext_body(self, session):
        group_list = bytearray()
        for group in self.supported_groups:
            group_list.extend(pdu.pack_uint16(getattr(group, "value", group)))

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint16(len(group_list)))
        ext_body.extend(group_list)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        length, offset = pdu.unpack_uint16(ext_body, 0)
        end_of_list = offset + length
        while offset < end_of_list:
            group, offset = pdu.unpack_uint16(ext_body, offset)
            try:
                group = tls.SupportedGroups(group)

            except ValueError:
                pass

            self.supported_groups.append(group)


class ExtSignatureAlgorithms(Extension):
    """Represents the SignatureAlgorithms extension.

    Attributes:
        signature_algorithms: The of supported signature algorithms.
    """

    extension_id = tls.Extension.SIGNATURE_ALGORITHMS
    """:obj:`tlsmate.tls.Extension.SIGNATURE_ALGORITHMS`
    """

    def __init__(
        self,
        signature_algorithms: Optional[List[Union[tls.SignatureScheme, int]]] = None,
    ) -> None:
        self.signature_algorithms = signature_algorithms if signature_algorithms else []

    def _serialize_ext_body(self, session):
        algo_list = bytearray()
        for algo in self.signature_algorithms:
            if type(algo) is tuple:
                algo = 256 * algo[0] + algo[1]
            algo_list.extend(pdu.pack_uint16(getattr(algo, "value", algo)))

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint16(len(algo_list)))
        ext_body.extend(algo_list)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        length, offset = pdu.unpack_uint16(ext_body, 0)
        end_of_list = offset + length
        while offset < end_of_list:
            algo, offset = pdu.unpack_uint16(ext_body, offset)
            try:
                algo = tls.SignatureScheme(algo)

            except ValueError:
                pass

            self.signature_algorithms.append(algo)


class ExtHeartbeat(Extension):
    """Represents the Heartbeat extension.

    Attributes:
        heartbeat_mode: The mode for the heartbeat extension.
    """

    extension_id = tls.Extension.HEARTBEAT
    """:obj:`tlsmate.tls.Extension.HEARTBEAT`
    """

    def __init__(self, heartbeat_mode: Optional[tls.HeartbeatMode] = None) -> None:
        self.heartbeat_mode = heartbeat_mode

    def _serialize_ext_body(self, session):
        if type(self.heartbeat_mode) is int:
            val = self.heartbeat_mode
        else:
            val = self.heartbeat_mode.value

        return pdu.pack_uint8(val)

    def _deserialize_ext_body(self, ext_body):
        mode, offset = pdu.unpack_uint8(ext_body, 0)
        self.heartbeat_mode = tls.HeartbeatMode.val2enum(mode, alert_on_failure=True)


class ExtCertificateAuthorities(Extension):
    """Represents the CertificateAuthorities extension.

    Attributes:
        authorities: The list of authorities in original ASN.1 format.
    """

    extension_id = tls.Extension.CERTIFICATE_AUTHORITIES
    """:obj:`tlsmate.tls.Extension.CERTIFICATE_AUTHORITIES`
    """

    def __init__(self, authorities: Optional[List[bytes]] = None) -> None:
        self.authorities = authorities if authorities else []

    def _serialize_ext_body(self, session):
        raise NotImplementedError(f"serialization of extension {self} not implemented")

    def _deserialize_ext_body(self, ext_body):
        length, offset = pdu.unpack_uint16(ext_body, 0)
        end_of_list = offset + length
        while offset < end_of_list:
            length, offset = pdu.unpack_uint16(ext_body, offset)
            authority, offset = pdu.unpack_bytes(ext_body, offset, length)
            self.authorities.append(authority)


class ExtSessionTicket(Extension):
    """Represents the SessionTicket extension.

    Attributes:
        ticket: The ticket.
    """

    extension_id = tls.Extension.SESSION_TICKET
    """:obj:`tlsmate.tls.Extension.SESSION_TICKET`
    """

    def __init__(self, ticket: Optional[bytes] = None) -> None:
        self.ticket = ticket

    def _serialize_ext_body(self, session):
        ext_body = bytearray()
        if self.ticket is not None:
            ext_body.extend(self.ticket)

        return ext_body

    def _deserialize_ext_body(self, ext_body):
        pass


class ExtStatusRequest(Extension):
    """Represents the status_request extension.

    Attributes:
        status_type: The status type. Defaults to
            :obj:`tlsmate.tls.StatusType.OCSP`.
        responder_ids: The list of responder ids. Defaults to an empty list.
        extensions: The list of extensions. Defaults to an empty list.
    """

    extension_id = tls.Extension.STATUS_REQUEST
    """:obj:`tlsmate.tls.Extension.STATUS_REQUEST`
    """

    def __init__(
        self,
        status_type: tls.StatusType = tls.StatusType.OCSP,
        responder_ids: Optional[List[bytes]] = None,
        extensions: Optional[List[bytes]] = None,
    ) -> None:
        self.status_type = status_type
        if responder_ids is None:
            responder_ids = []

        self.reponder_ids = responder_ids
        if extensions is None:
            extensions = []

        self.extensions = extensions
        self.ocsp_response = None

    def _serialize_ext_body(self, session):
        ext_body = bytearray()
        ext_body.extend(
            pdu.pack_uint8(getattr(self.status_type, "value", self.status_type))
        )

        responders = bytearray()
        for responder in self.reponder_ids:
            responders.extend(pdu.pack_uint16(len(responder)))
            responders.extend(responder)

        ext_body.extend(pdu.pack_uint16(len(responders)))
        ext_body.extend(responders)

        extensions = bytearray()
        for extension in self.extensions:
            extensions.extend(pdu.pack_uint16(len(extension)))
            extensions.extend(extension)

        ext_body.extend(pdu.pack_uint16(len(extensions)))
        ext_body.extend(extensions)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        if not len(ext_body):
            self.status_type = tls.StatusType.NONE
            return

        status_type, offset = pdu.unpack_uint8(ext_body, 0)
        self.status_type = tls.StatusType.val2enum(status_type)
        length, offset = pdu.unpack_uint24(ext_body, offset)
        self.ocsp_response, offset = pdu.unpack_bytes(ext_body, offset, length)


class ExtStatusRequestV2(Extension):
    """Represents the status_requestv2 extension.

    Attributes:
        status_type (:obj:`tlsmate.tls.StatusType`): The status type. Defaults to
            :obj:`tlsmate.tls.StatusType.OCSP`.
        responder_ids (list of bytes): The list of responder ids. Defaults to an empty
            list.
        extensions (list of bytes): The list of extensions. Defaults to an empty
            list.
    """

    extension_id = tls.Extension.STATUS_REQUEST_V2
    """:obj:`tlsmate.tls.Extension.STATUS_REQUEST_V2`
    """

    def __init__(
        self,
        status_type: tls.StatusType = tls.StatusType.OCSP_MULTI,
        responder_ids: Optional[List[bytes]] = None,
        extensions: bytes = b"",
    ) -> None:
        self._requests: List[Tuple[tls.StatusType, Optional[List[bytes]], bytes]] = []
        self.add_request(status_type, responder_ids, extensions)

    def add_request(self, status_type, responder_ids=None, extensions=b""):
        if responder_ids is None:
            responder_ids = []

        self._requests.append((status_type, responder_ids, extensions))

    def _serialize_ext_body(self, session):

        ext_body = bytearray()
        for status_type, responder_ids, extensions in self._requests:
            request_item = bytearray(
                pdu.pack_uint8(getattr(status_type, "value", status_type))
            )
            status_request = bytearray()
            responders = bytearray()
            for responder in responder_ids:
                responders.extend(pdu.pack_uint16(len(responder)))
                responders.extend(responder)
            status_request.extend(pdu.pack_uint16(len(responders)))
            status_request.extend(responders)
            status_request.extend(pdu.pack_uint16(len(extensions)))
            status_request.extend(extensions)
            request_item.extend(pdu.pack_uint16(len(status_request)))
            request_item.extend(status_request)
            ext_body.extend(pdu.pack_uint16(len(request_item)))
            ext_body.extend(request_item)

        return ext_body

    def _deserialize_ext_body(self, ext_body):
        if not len(ext_body):
            self.status_type = tls.StatusType.NONE
            return

        status_type, offset = pdu.unpack_uint8(ext_body, 0)
        self.status_type = tls.StatusType.val2enum(status_type)
        length, offset = pdu.unpack_uint24(ext_body, offset)
        self.ocsp_response, offset = pdu.unpack_bytes(ext_body, offset, length)


class ExtSupportedVersions(Extension):
    """Represents the SupportedVersion extension.

    Attributes:
        versions: The list of TLS versions supported.
    """

    extension_id = tls.Extension.SUPPORTED_VERSIONS
    """:obj:`tlsmate.tls.Extension.SUPPORTED_VERSIONS`
    """

    def __init__(
        self, versions: Optional[List[Union[tls.Version, int]]] = None
    ) -> None:
        self.versions = versions

    def _serialize_ext_body(self, session):
        versions = bytearray()
        for version in self.versions:
            versions.extend(pdu.pack_uint16(getattr(version, "value", version)))

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint8(len(versions)))
        ext_body.extend(versions)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        self.versions = []
        offset = 0
        while offset < len(ext_body):
            version, offset = pdu.unpack_uint16(ext_body, offset)
            version = tls.Version.val2enum(version)
            self.versions.append(version)


class ExtCookie(Extension):
    """Represents the Cookie extension.

    Attributes:
        cookie: The opaque cookie value
    """

    extension_id = tls.Extension.COOKIE
    """:obj:`tlsmate.tls.Extension.COOKIE`
    """

    def __init__(self, cookie: bytes = b"") -> None:
        self.cookie = cookie

    def _serialize_ext_body(self, session):
        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint16(len(self.cookie)))
        ext_body.extend(self.cookie)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        length, offset = pdu.unpack_uint16(ext_body, 0)
        self.cookie, offset = pdu.unpack_bytes(ext_body, offset, length)


class ExtKeyShare(Extension):
    """Represents the KeyShare extension.

    Given the list of supported groups, during serialization the shares are actually
    generated and the resulting public keys are included into the extension. The user
    must ensure that the key share uses the same sequence for the groups than
    provided in the SupportedGroup extension.

    Attributes:
        key_shares (list of :obj:`tlsmate.structs.KeyShareEntry`): the list of
            key shares, comprising the group and the public key for each element.
        groups (list of :obj:`tlsmate.tls.SupportedGroups`): The list of
            supported groups for which key shares shall be generated. This is an
            alternative to setup the extension. The public key will be determined
            when serializing the extension.
    """

    extension_id = tls.Extension.KEY_SHARE
    """:obj:`tlsmate.tls.Extension.KEY_SHARE`
    """

    def __init__(
        self,
        key_shares: Optional[List[structs.KeyShareEntry]] = None,
        groups: Optional[List[tls.SupportedGroups]] = None,
    ) -> None:
        if key_shares is None:
            key_shares = []

        if groups is None:
            groups = []

        if not key_shares and groups:
            key_shares = [
                structs.KeyShareEntry(group=group, key_exchange=None)
                for group in groups
            ]

        self.key_shares = key_shares

    def _serialize_ext_body(self, session):
        key_shares = bytearray()
        for ks in self.key_shares:
            key_shares.extend(pdu.pack_uint16(ks.group.value))
            share = session.create_key_share(ks.group)
            key_shares.extend(pdu.pack_uint16(len(share)))
            key_shares.extend(share)

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint16(len(key_shares)))
        ext_body.extend(key_shares)
        return ext_body

    def _deserialize_ext_body(self, ext_body):
        self.key_shares = []
        offset = 0
        while offset < len(ext_body):
            group, offset = pdu.unpack_uint16(ext_body, offset)
            group = tls.SupportedGroups.val2enum(group, alert_on_failure=True)
            if offset < len(ext_body):
                length, offset = pdu.unpack_uint16(ext_body, offset)
                share, offset = pdu.unpack_bytes(ext_body, offset, length)

            else:
                share = None

            self.key_shares.append(
                structs.KeyShareEntry(group=group, key_exchange=share)
            )


class ExtPreSharedKey(Extension):
    """Represents the PreSharedKey extension for TLS1.3

    Builds this extension based on the list of the given pre shared keys.

    Attributes:
        psks: The list of pre shared keys to offer to the server.
    """

    extension_id = tls.Extension.PRE_SHARED_KEY
    """:obj:`tlsmate.tls.Extension.PRE_SHARED_KEY`
    """

    def __init__(self, psks: Optional[List[structs.Psk]] = None) -> None:
        self.psks = psks
        self._bytes_after_ids = 0

    def _serialize_ext_body(self, session):
        identities = bytearray()
        binders = bytearray()
        for psk in self.psks:
            identities.extend(pdu.pack_uint16(len(psk.ticket)))
            identities.extend(psk.ticket)
            timestamp = session.recorder.inject(timestamp=time.time())
            ticket_age = int((timestamp - psk.timestamp) * 1000 + psk.age_add) % (
                2 ** 32
            )
            identities.extend(pdu.pack_uint32(ticket_age))
            binders.extend(pdu.pack_uint8(psk.hmac.mac_len))
            binders.extend(b"\0" * psk.hmac.mac_len)

        ext_body = bytearray()
        ext_body.extend(pdu.pack_uint16(len(identities)))
        ext_body.extend(identities)
        binders_len = len(binders)
        self._bytes_after_ids = binders_len + 2  # 2 bytes for length indicator
        ext_body.extend(pdu.pack_uint16(binders_len))
        ext_body.extend(binders)

        return ext_body

    def _deserialize_ext_body(self, ext_body):
        self.selected_id, offset = pdu.unpack_uint16(ext_body, 0)


class ExtPskKeyExchangeMode(Extension):
    """Represents the psk_key_exchange_mode extension.

    Attributes:
        modes: The list of the PSK key exchange modes to offer to the server.
    """

    extension_id = tls.Extension.PSK_KEY_EXCHANGE_MODES
    """:obj:`tlsmate.tls.Extension.PSK_KEY_EXCHANGE_MODES`
    """

    def __init__(self, modes: Optional[List[tls.PskKeyExchangeMode]] = None) -> None:
        self.modes = modes

    def _serialize_ext_body(self, session):
        ext_body = bytearray(pdu.pack_uint8(len(self.modes)))
        for mode in self.modes:
            ext_body.extend(pdu.pack_uint8(getattr(mode, "value", mode)))

        return ext_body

    def _deserialize_ext_body(self, ext_body):
        raise NotImplementedError


class ExtEarlyData(Extension):
    """Represents the EarlyData extension.

    Attributes:
        max_early_data_size: The maximum number of bytes used for the early data.
    """

    extension_id = tls.Extension.EARLY_DATA
    """:obj:`tlsmate.tls.Extension.EARLY_DATA`
    """

    def __init__(self, max_early_data_size: Optional[int] = None) -> None:
        self.max_early_data_size = max_early_data_size

    def _serialize_ext_body(self, session):
        if self.max_early_data_size is None:
            return b""

        else:
            return bytes(pdu.pack_uint32(self.max_early_data_size))

    def _deserialize_ext_body(self, ext_body):
        if ext_body:
            self.max_early_data_size, _ = pdu.unpack_uint32(ext_body, 0)


class ExtPostHandshakeAuth(Extension):
    """Represents the PostHandshakeAuth extension.
    """

    extension_id = tls.Extension.POST_HANDSHAKE_AUTH
    """:obj:`tlsmate.tls.Extension.POST_HANDSHAKE_AUTH`
    """

    def _serialize_ext_body(self, session):
        return b""

    def _deserialize_ext_body(self, ext_body):
        if ext_body:
            raise tls.ServerMalfunction(
                tls.ServerIssue.EXTENTION_LENGHT_ERROR, extension=self.extension_id
            )


# Map the extensions id to the corresponding class.
deserialization_map = {
    tls.Extension.SERVER_NAME: ExtServerNameIndication,
    # tls.Extension.MAX_FRAGMENT_LENGTH = 1
    # tls.Extension.CLIENT_CERTIFICATE_URL = 2
    # tls.Extension.TRUSTED_CA_KEYS = 3
    # tls.Extension.TRUNCATED_HMAC = 4
    tls.Extension.STATUS_REQUEST: ExtStatusRequest,
    # tls.Extension.USER_MAPPING = 6
    # tls.Extension.CLIENT_AUTHZ = 7
    # tls.Extension.SERVER_AUTHZ = 8
    # tls.Extension.CERT_TYPE = 9
    tls.Extension.SUPPORTED_GROUPS: ExtSupportedGroups,
    tls.Extension.EC_POINT_FORMATS: ExtEcPointFormats,
    # tls.Extension.SRP = 12
    tls.Extension.SIGNATURE_ALGORITHMS: ExtSignatureAlgorithms,
    # tls.Extension.USE_SRTP = 14
    tls.Extension.HEARTBEAT: ExtHeartbeat,
    # tls.Extension.APPLICATION_LAYER_PROTOCOL_NEGOTIATION = 16
    tls.Extension.STATUS_REQUEST_V2: ExtStatusRequestV2,
    # tls.Extension.SIGNED_CERTIFICATE_TIMESTAMP = 18
    # tls.Extension.CLIENT_CERTIFICATE_TYPE = 19
    # tls.Extension.SERVER_CERTIFICATE_TYPE = 20
    # tls.Extension.PADDING = 21
    tls.Extension.ENCRYPT_THEN_MAC: ExtEncryptThenMac,
    tls.Extension.EXTENDED_MASTER_SECRET: ExtExtendedMasterSecret,
    # tls.Extension.TOKEN_BINDING = 24
    # tls.Extension.CACHED_INFO = 25
    # tls.Extension.TLS_LTS = 26
    # tls.Extension.COMPRESS_CERTIFICATE = 27
    # tls.Extension.RECORD_SIZE_LIMIT = 28
    # tls.Extension.PWD_PROTECT = 29
    # tls.Extension.PWD_CLEAR = 30
    # tls.Extension.PASSWORD_SALT = 31
    # tls.Extension.TICKET_PINNING = 32
    # tls.Extension.TLS_CERT_WITH_EXTERN_PSK = 33
    # tls.Extension.DELEGATED_CREDENTIALS = 34
    tls.Extension.SESSION_TICKET: ExtSessionTicket,
    # tls.Extension.SUPPORTED_EKT_CIPHERS = 39
    tls.Extension.PRE_SHARED_KEY: ExtPreSharedKey,
    tls.Extension.EARLY_DATA: ExtEarlyData,
    tls.Extension.SUPPORTED_VERSIONS: ExtSupportedVersions,
    tls.Extension.COOKIE: ExtCookie,
    # tls.Extension.PSK_KEY_EXCHANGE_MODES = 45
    tls.Extension.CERTIFICATE_AUTHORITIES: ExtCertificateAuthorities,
    # tls.Extension.OID_FILTERS = 48
    tls.Extension.POST_HANDSHAKE_AUTH: ExtPostHandshakeAuth,
    # tls.Extension.SIGNATURE_ALGORITHMS_CERT = 50
    tls.Extension.KEY_SHARE: ExtKeyShare,
    # tls.Extension.TRANSPARENCY_INFO = 52
    # tls.Extension.CONNECTION_ID = 53
    # tls.Extension.EXTERNAL_ID_HASH = 55
    # tls.Extension.EXTERNAL_SESSION_ID = 56
    tls.Extension.RENEGOTIATION_INFO: ExtRenegotiationInfo,
}


def log_extensions(extensions: List[Extension]) -> None:
    """Log extensions

    Arguments:
        extensions: the list of extensions to iterate over
    """

    for extension in extensions:
        ext_id = extension.extension_id
        logging.debug(f"extension {ext_id.value} {ext_id}")


def get_extension(
    extensions: List[Extension], ext_id: tls.Extension
) -> Optional[Extension]:
    """Helper function to search for an extension

    Arguments:
        extensions: the list to search for the extension
        ext_id: the extension to to look for

    Returns:
        The extension if present, or None otherwise.
    """

    if extensions is not None:
        for extension in extensions:
            if extension.extension_id == ext_id:
                return extension

    return None


def deserialize_extensions(
    extensions: List[Extension], fragment: bytes, offset: int
) -> int:
    """Helper function to deserialize extensions

    Arguments:
        extensions: the list where to store the deserialized extensions
        fragment: the pdu buffer
        offset: the offset within the pdu buffer to the start of the extensions

    Returns:
        the offset on the byte following the extensions
    """

    if offset < len(fragment):
        # extensions present
        ext_len, offset = pdu.unpack_uint16(fragment, offset)
        ext_end = offset + ext_len
        while offset < ext_end:
            extension, offset = Extension.deserialize(fragment, offset)
            extensions.append(extension)

    return offset



# -*- coding: utf-8 -*-
"""Implements a class to be used for unit testing.
"""
import datetime
from tlsmate.cert_chain import CertChain
from tlsmate.server_profile import SPObject, ProfileSchema, ServerProfileSchema
from tlsmate import utils

from marshmallow import fields
import pytest


class SPUnitTest(SPObject):
    pass


class SPUnitTestSchema(ProfileSchema):
    __profile_class__ = SPUnitTest

    unit_test_1 = fields.Integer()
    unit_test_2 = fields.String()


@ServerProfileSchema.augment
class SPUnitTestAugment(ProfileSchema):
    unit_test = fields.Nested(SPUnitTestSchema)


def test_cert_paras(tlsmate, guballa_de_pem, quo_vadis_root_ca3):
    chain = CertChain()
    for cert in (quo_vadis_root_ca3, guballa_de_pem):
        chain.append_pem_cert(cert.as_bytes())

    prof = tlsmate.server_profile
    prof.allocate_versions()
    prof.append_unique_cert_chain(chain)

    quo_vadis = prof.cert_chains[0].cert_chain[0]
    guballa = prof.cert_chains[0].cert_chain[1]

    cert_policies = quo_vadis.extensions[1].cert_policies
    explicit_text = cert_policies[0].policy_qualifiers[0].explicit_text
    assert type(explicit_text) is str
    assert len(explicit_text)

    text = cert_policies[0].policy_qualifiers[1].text
    assert type(text) is str
    assert len(text)

    signed_ct = guballa.extensions[8].signed_certificate_timestamps
    assert len(signed_ct) == 2
    for ct in signed_ct:
        assert ct.entry_type == "PRE_CERTIFICATE"
        assert type(ct.log_id) is bytes
        assert len(ct.log_id)
        assert type(ct.timestamp) is datetime.datetime
        assert ct.version == "v1"


def test_augment_profile(tlsmate):
    tlsmate.server_profile.unit_test = SPUnitTest(unit_test_1=1, unit_test_2="hello")
    data = tlsmate.server_profile.make_serializable()
    assert "unit_test" in data
    assert data["unit_test"]["unit_test_1"] == 1
    assert data["unit_test"]["unit_test_2"] == "hello"


def test_deserialize_profile_ok(tlsmate):
    data = {"unit_test": {"unit_test_1": 1, "unit_test_2": "hello"}}
    tlsmate.server_profile.load(data)
    assert tlsmate.server_profile.unit_test.unit_test_1 == 1
    assert tlsmate.server_profile.unit_test.unit_test_2 == "hello"


def test_deserialize_profile_nok(tlsmate):
    data = {
        "unit_test": {"unit_test_1": 1, "unit_test_2": "hello"},
        "too_much": "outch",
    }

    with pytest.raises(ValueError, match="fields not defined in schema"):
        tlsmate.server_profile.load(data)


def test_deserialize_full_profile(tlsmate, server_profile):
    tlsmate.server_profile.load(utils.deserialize_data(server_profile))
